#######################################################################################
# This file contains R code to process root phenotype data from the plate rotation
# experiments of the tomato M82/PENN introgression lines done by Edgar Spalding's
# lab.  Those were time course experiments that produced 34 time points over a
# course of two hours.  There are two input data files:
#       allILtipangles.csv: tip angle at each of 34 time points for each sample,
#           where each sample is some IL or parent.
#       allILderiv.csv: derivative of the tip angle with respect to time, 34 time
#           points per sample.
# The "date" column in these files is bogus and should be ignored.
#
# This file runs a loop to process each of the two input files.  It also runs a loop
# to process each file three different times, using these different methods of
# selecting the data to be processed:
#   "THRU" - process data for roots rotated THROUGH the gravity vector
#   "AWAY" - process data for roots rotated AWAY from the gravity vector
#   "ALL" - process data for both rotations
#
# Output filenames are different for each of the two input files and for each of
# the three data selection methods shown above.
#
# To run this code, select all the code in the file and execute it.  Check the
# console or output (recorded in the output log file) for errors.
#
# Revision history:
# Author    Date        Description
# --------  -------     ---------------------------------------------------------------
# Ted Toal  22Jan15     Create this file from a copy of the Mily RootLength_Gravity.r file
#                       that had been modified by Edgar Spalding.  I edited it further.
#                       Date was removed, there was no recorded experiment date.  Instead
#                       of lme(), which used date as a random effect, we now use lm().
# Ted Toal  12Feb15     Change trait names.  I need the traits in the derivative version
#                       to have different names than in the straight angle version.
# Ted Toal  Mar2015     Split plotting code out into new file.  Also made several
#                       changes that were previously made in the RootRNAseqJune2013 code.
# Ted Toal  20Apr2015   Add rotMethod and loop.
# Ted Toal  20May2015   Change groups.
# Ted Toal  24Jul2015   Add alignAnglesAtT0.
#######################################################################################

# Include the root R source code file.
R_INCLUDE_DIR = Sys.getenv("R_INCLUDE_DIR")
if (R_INCLUDE_DIR == "")
    stop("Set environment variable R_INCLUDE_DIR or assign that variable to a path here.")
source(paste(R_INCLUDE_DIR, "Include_RootRsourceFile.R", sep=""))

# Source our basic utility functions file.
source(Include_UtilityFunctions)

# Load libraries we need.
addThisLibrary("xlsx")

# Source other files we need.
source(Include_TextFunctions)
source(Include_xlsx)
source(Include_TomatoILsBILs)
source(Include_phenoRNAseqRootSep2011)
source(Include_phenoAnalysis)

# Set working directory.
setwd(SpaldingAnalysisDir)

# Significant p-values.
sigp = c(0.05, 0.01, 0.001)

# Set this TRUE to force all tip angles of each genotype to be altered by subtracting
# the angle at time 0, giving all genotypes a tip angle of 0 at time 0.  This is so
# that a reasonable tip angle comparison can be done.  In this case, trait T0 is
# removed.
alignAnglesAtT0 = TRUE

# SET THIS TO TODAY'S DATE, the date of reading and plotting data.  This date
# prefixes all input and all output filenames AND IS OUTPUT DIRECTORY NAME.
# If you don't change this, you may overwrite files already generated that
# are to be kept.
fileDate = "21May2015"

# SET THIS TO A NAME TO FOLLOW fileDate IN THE OUTPUT FOLDER NAME TO IDENTIFY
# WHAT DISTINGUISHES THIS RUN FROM OTHERS.
{
if (!alignAnglesAtT0)
    folderSfx = "_AwayThru"
else
    folderSfx = "_AlignAtT0"
}

# Input filenames.
inputFilename_swingRate = "allILderiv.csv"
inputFilename_tipAngle = "allILtipangles.csv"

# Output folder name.
outputFolder = paste(fileDate, folderSfx, sep="")

# Create the output folder if it doesn't exist.
dir.create(outputFolder, showWarnings=FALSE)

# Column names to be removed from input data.
removeCols = "date"

# Define a function to read an input file and fix up some problems with it,
# returning the resulting data frame.
readInputFile = function(inputFilename)
    {
    data = read.csv(inputFilename, stringsAsFactors=FALSE)
    #str(data)
    data = gdata::trim(data) # Get rid of leading and trailing spaces.
    data$genotype = toupper(data$genotype) # Use UPPER CASE for all alphabetic characters
    data$genotype = gsub("--", "-", data$genotype, fixed=TRUE)  # Use just one "-" in IL names.
    data$genotype = gsub(" ", "", data$genotype, fixed=TRUE)  # No spaces in genotype names.
    data$genotype = paste("IL", data$genotype, sep="") # Make sure every IL starts with "IL" (except M82 and PENN).
    data$genotype = gsub("^ILIL", "IL", data$genotype)
    data$genotype = gsub("^ILM82", "M82", data$genotype)
    data$genotype = gsub("^ILPENN", "PENN", data$genotype)
    # Fix IL7-5-P5.
    data$genotype = sub("ILP5", "IL7-5-P5", data$genotype)
    # Fix IL8-3-1.
    data$genotype = sub("IL8-3-1A", "IL8-3-1", data$genotype)
    # Make genotype a factor with M82 as the reference level.
    data$genotype = factor(data$genotype)
    data$genotype = relevel(data$genotype, ref="M82")
    # Remove columns that we don't want.
    data = data[, -which(colnames(data) %in% removeCols)]
    # Rename the columns from T1..T34 to T0..T33, since the first time point is
    # at time point 0.  It makes more sense.  (But maybe we shouldn't.  Oh, what
    # the heck.)  If alignAnglesAtT0 is true, the T0 column will later be removed.
    colnames(data)[2:ncol(data)] = paste("T", 0:(ncol(data)-2), sep="")
    return(data)
    }

# Read the tip angle file into dfTipAngle, we will need it below even when we
# are processing derivative data to determine whether each sample is AWAY or
# THRU (and therefore this data frame is specifically NOT aligned even when
# alignAnglesAtT0 is TRUE). 
dfTipAngle = readInputFile(inputFilename_tipAngle)

################################################################################
################################################################################
################################################################################
# Loop for each rotation method:
for (rotMethod in c("THRU", "AWAY", "ALL"))
    {
    # Set criteria for data filtering.
    {
    if (rotMethod == "ALL")
        minRootsPerIL = 4
    else
        minRootsPerIL = 3
    }
    criteriaCode = paste("min", minRootsPerIL, sep="")

    # Loop for both input files (tip angles and derivatives).
    for (analyzeDeriv in c(FALSE, TRUE))
        {
        # Input filename.
        {
        if (analyzeDeriv)
            inputFilename = inputFilename_swingRate
        else
            inputFilename = inputFilename_tipAngle
        }

        # Set output filename prefix.
        rootOutputFilePrefix = paste(fileDate, "_", sub(".csv", "", inputFilename), "_",
            rotMethod, "_", criteriaCode, "_", sep="")
        rootOutputFilePrefix = paste(outputFolder, rootOutputFilePrefix, sep=PATHSEP)

        # Make empty log file.
        logfile = paste(rootOutputFilePrefix, "logfile.txt", sep="")
        setDefaultLogFile(logfile)
        printLog("Spalding lab plate rotation experiment\n", append=FALSE)
        printLog("analyzeDeriv: ", analyzeDeriv, "\n", sep="")
        printLog("rotMethod: ", rotMethod, "\n", sep="")
        printLog("Root angle analysis log file: ", logfile, "\n")

        ##################################
        # Input data file and traits.
        ##################################

        # Read the data file.
        data = readInputFile(inputFilename)

        # Get trait names and number of traits.
        traits = colnames(data)[2:ncol(data)]
        if (alignAnglesAtT0 && !analyzeDeriv)
            traits = setdiff(traits, "T0")
        ntraits = length(traits)
        printLog("Data summary:\n")
        printLog(getPrint(summary(data)))
        printLog("Column names:\n")
        printLog(colnames(data), "\n")
        printLog("There are", ntraits, "traits in total: ", traits, "\n")

        # Get genotypes and number of genotypes.
        genotypes = levels(data$genotype)
        ngenotypes = length(genotypes) #how long is the vector (140 #genotypes)
        printLog("There are", ngenotypes, "genotypes in total:\n")
        printLog(genotypes, "\n", fill=70, labels="   ")

        # Set a trait description vector.
        {
        if (alignAnglesAtT0 && !analyzeDeriv)
            traitDescs = paste("T", 1:length(traits), sep="")
        else
            traitDescs = paste("T", 0:(length(traits)-1), sep="")

        if (rotMethod == "THRU")
            traitDescs = paste("Thru ", traitDescs, sep="")
        else if (rotMethod == "AWAY")
            traitDescs = paste("Away ", traitDescs, sep="")

        if (analyzeDeriv)
            traitDescs = paste("Angle Rate ", traitDescs, sep="")
        else
            traitDescs = paste("Angle ", traitDescs, sep="")
        }
        names(traitDescs) = traits
        printLog("Trait descriptions:\n")
        printLog(traitDescs, "\n")

        # Set a trait units vector.
        {
        if (analyzeDeriv)
            traitUnits = rep("deg/225s", length(traits))
        else
            traitUnits = rep("degrees", length(traits))
        }
        names(traitUnits) = traits
        printLog("Trait units of measurement:\n")
        printLog(traitUnits, "\n")

        # Make a data frame containing columns "unit" and "desc", with rownames equal to
        # trait names, and save it in the output folder.
        dfTraits = data.frame(units=traitUnits, desc=traitDescs, row.names=traits, stringsAsFactors=FALSE)
        filename = paste(rootOutputFilePrefix, "traits.tsv", sep="")
        write.table(dfTraits, filename, sep="\t", row.names=TRUE, col.names=TRUE)
        # dfTraits = read.table(filename, sep="\t", header=TRUE, row.names=1, stringsAsFactors=FALSE)

        ##################################
        # Groups of traits.
        ##################################

        # Define groups of similar or related traits.  These trait groups might be used to
        # group together genes D.E. in any IL sig. different than M82 for any trait in the
        # group, in order to analyze the genes for enrichments or coexpression that would
        # provide clues to genes driving the traits.

        # Trait groups are defined with a special formula.  The description of the formula
        # "language" is located in the comments for the function getILsInGroups() in file
        # Include_phenoAnalysis.R.

        # Although the formula language permits great expressiveness in defining groups,
        # each group should be defined according to some expected biological significance.

        # Here we group traits, which in this case are time points, together into early,
        # middle, and late response time points, based on the heat maps produced above.

        # Group names MAY be the same as trait names.

        # Define groups by defining a vector whose element names are the group names, and
        # whose element values are the group formulas.  Note that we could define
        # separate groups for Deriv vs not, and for each rotMethod ALL, AWAY, and
        # THRU.  I'm going to try NOT doing this but using the same groups for all,
        # which will be much easier and make more sense really.

        traitGroups = c(
            Early = ifelse(alignAnglesAtT0 && !analyzeDeriv, "T1|T2|T3|T4|T5|T6", "T0|T1|T2|T3|T4|T5|T6"), # 22.5 minutes
            Middle = "T7|T8|T9|T10|T11|T12|T13|T14|T15|T16|T17|T18|T19", # 49 minutes
            Late = "T20|T21|T22|T23|T24|T25|T26|T27|T28|T29|T30|T31|T32|T33", # 52.5 minutes
            NULL) # NULL doesn't get entered into the vector.

        dfGroups = data.frame(group=names(traitGroups), formula=traitGroups,
            stringsAsFactors=FALSE)
        filename = paste(rootOutputFilePrefix, "traitGroups.tsv", sep="")
        write.table(dfGroups, filename, sep="\t", row.names=FALSE, col.names=TRUE)

        ##################################
        # Each trait in this file is actually a time point.  Remove data points for
        # roots whose rotation does not conform to the setting of rotMethod.  All
        # traits (time points) have exactly the same data available for them.
        # Remove ILs with NA data points, there shouldn't actually be any.  Then
        # make sure the data includes at least "minRootsPerIL" roots for each IL.
        # Any IL that has fewer than "minRootsPerIL" roots is discarded.  All ILs
        # containing one or more NA samples are discarded (there are none).
        # Whatever is left is stored in a new data frame with these columns:
        # genotype, trait.  "data.all" holds the resulting data frame.
        printLog("Reducing data set to include only those genotypes with at least", minRootsPerIL, "roots.\n")
        ##################################

        # Copy the input data frame.
        data.all = data

        # Remove roots whose rotation doesn't conform to rotMethod setting.
        {
        Nr = nrow(data.all)
        if (rotMethod == "THRU")
            data.all = data.all[dfTipAngle[,2] >= 0,]
        else if (rotMethod == "AWAY")
            data.all = data.all[dfTipAngle[,2] < 0,]
        Nrmv = Nr - nrow(data.all)
        if (Nrmv > 0)
            printLog("Removed ", Nrmv, " roots not rotated ", rotMethod, "\n", sep="")
        }

        # Remove ILs with NA data points.
        hasNA = apply(data.all[, -1], 1, function(V) any(is.na(V)))
        names(hasNA) = data.all[, 1]
        if (any(hasNA))
            {
            rmvILs = names(hasNA)[hasNA]
            data.all = data.all[!data.all[,1] %in% rmvILs,]
            printLog("Removed ILs with NA data: ", paste(rmvILs, collapse=","), "\n", sep="")
            }

        # Further reduce the data set to only those ILs that have at least
        # minRootsPerIL roots.
        numRoots = tapply(1:nrow(data.all), data.all$genotype, length)
        numRoots[is.na(numRoots)] = 0
        keep = (numRoots[data.all$genotype] >= minRootsPerIL)
        Nr = nrow(data.all)
        rmvILs = unique(data.all$genotype[!keep])
        data.all = data.all[keep,]
        Nrmv = Nr - nrow(data.all)
        if (Nrmv > 0)
            printLog("Removed ", length(rmvILs), " ILs containing ", Nrmv, " roots having fewer than ",
                minRootsPerIL, " roots per IL:\n   ", paste(rmvILs, collapse=","), "\n", sep="")

        # Get rid of unused levels.
        data.all$genotype = factor(data.all$genotype)

        # Count number of roots for each genotype (for each trait).
        numRoots = tapply(1:nrow(data.all), data.all$genotype, function(x) sum(!is.na(x)))
        numRoots = numRoots[!(rownames(numRoots) %in% parents)]
        printLog("   Minimum number of roots for any IL:", min(numRoots), "\n")
        printLog("   Maximum number of roots for any IL:", max(numRoots), "\n")
        printLog("   Mean number of roots for all ILs:", mean(numRoots), "\n")
        printLog("   SD for number of roots for all ILs:", sd(numRoots), "\n")

        # If no data is left, say so.
        if (nrow(data.all) == 0)
            {
            printLog("Trait", t, "has no samples that satisfy the minimal data criteria!!!\n")
            stop("No data")
            }

        # Set m82 as reference level.
        if (!any(data.all$genotype == "M82"))
            {
            printLog("Trait", t, "has no M82 samples that satisfy the minimal data criteria!!!\n")
            stop("No M82")
            }
        data.all$genotype = relevel(data.all$genotype, ref="M82")

        # Tell the user which genotypes were eliminated and which retained as a result of above.
        new.genotypes = levels(data.all$genotype)
        new.ngenotypes = length(new.genotypes)
        old.genotypes = setdiff(genotypes, new.genotypes)
        old.ngenotypes = length(old.genotypes)
        printLog("There are", new.ngenotypes, "genotypes remaining after removing those with no data:\n")
        printLog(new.genotypes, "\n", fill=70, labels="   ")
        printLog("There were", old.ngenotypes, "genotypes removed.  They were:\n")
        printLog(old.genotypes, "\n", fill=70, labels="   ")

        ##################################
        # Align tip angles at T0 if alignAnglesAtT0 is true and this is a tip angle dataset.
        ##################################
        if (alignAnglesAtT0 && !analyzeDeriv)
            {
            for (col in traits)
                data.all[, col] = data.all[, col] - data.all[, "T0"]
            data.all = data.all[, colnames(data.all) != "T0"]
            }

        ##################################
        # Write the reduced data frame to a file.
        ##################################
        vettedInputDataFile = paste(rootOutputFilePrefix, "inputDataFiltered.tsv", sep="")
        write.table(data.all, vettedInputDataFile, sep="\t", row.names=FALSE)

        ##################################
        # Make results matrices.  We will write data from these matrices to files after we compute them.
        ##################################
        model_matrix = matrix(nrow=ngenotypes, ncol=ntraits)
        rownames(model_matrix) = genotypes
        colnames(model_matrix) = traits
        #model_matrix

        r.N = model_matrix
        r.mean = model_matrix
        r.sd = model_matrix
        r.cv = model_matrix
        r.sem = model_matrix
        r.pvals = model_matrix
        r.qvals = model_matrix
        r.stars = model_matrix
        r.plus = model_matrix

        ##################################
        # Run loop for each trait.  Results will go into the above matrices.
        ##################################
        for (t in traits)
            {
            printLog("Processing trait: ", t, "\n")
            flush.console()

            # Get reduced data frame derived above and change name of the current trait column to "trait"
            data.r = data.all[, c("genotype", t)]
            colnames(data.r)[2] = "trait"

            # Fit the model.
            lm1 = lm(trait~genotype, data=data.r, na.action=na.omit)
            lm1.sum = summary(lm1)
    
            # Calculations for p-values in r.pvals.
            pvals = coef(lm1.sum)[, "Pr(>|t|)"]
            names(pvals) = sub("genotype", "", names(pvals)) #Strip Genotype from names so they will match
            pvals = pvals[-1] # Remove M82, meaningless p-value
            r.pvals[names(pvals), t] = pvals #aligns things correctly.  M82 gets NA but OK because p-value there was meaningless

            # Perform multiple hypothesis testing correction.
            qvals = p.adjust(pvals, method="fdr")
            r.qvals[names(qvals), t] = qvals
    
            # Calculate number of replicates for each genotype.
            numByGenotype = tapply(data.r$trait, data.r$genotype, function(V) sum(!is.na(V)))
            r.N[names(numByGenotype), t] = numByGenotype # Align things correctly

            # Siobhan's calculation's for means in r.means.
            meansByGenotype = tapply(data.r$trait, data.r$genotype, mean, na.rm=T)
            r.mean[names(meansByGenotype),t] = meansByGenotype # Align things correctly

            # Std deviations in r.sd.
            sdsByGenotype = tapply(data.r$trait, data.r$genotype, sd, na.rm=T)
            r.sd[names(sdsByGenotype),t] = sdsByGenotype # Align things correctly

            # Calculate coefficients of variation.
            CVsByGenotype = tapply(data.r$trait, data.r$genotype, function(x) sd(x, na.rm=T)/mean(x, na.rm=T))
            r.cv[names(CVsByGenotype), t] = CVsByGenotype # Align things correctly

            # Set the std. error values aside separately.
            sem = coef(lm1.sum)[, "Std. Error"]
            names(sem) = sub("genotype", "", names(sem)) #Strip Genotype from names so they will match
            names(sem)[1] = "M82" # Change "(Intercept)" to "M82"
            r.sem[names(sem), t] = sem

            # Make a stars representation of the corrected p-values.
            pvalGenotypes = names(qvals)
            r.stars[,t] = ""
            for (i in 1:length(sigp))
                r.stars[pvalGenotypes[qvals < sigp[i]], t] = paste(rep("*",i), collapse="")

            # Make a plus-sign representation of the unadjusted p-values.
            pvalGenotypes = names(pvals)
            r.plus[,t] = ""
            for (i in 1:length(sigp))
                r.plus[pvalGenotypes[pvals < sigp[i]], t] = paste(rep("+",i), collapse="")
            }
        #r.N
        #r.mean
        #r.sd
        #r.cv
        #r.sem
        #r.pvals
        #r.qvals
        #r.stars
        #r.plus

        # Write matrices of information about the model fits to files.
        # col.names=NA will leave a blank cell in the name of the Genotypes column
        write.table(r.N, paste(rootOutputFilePrefix, "N.tsv", sep=""), sep="\t", col.names=NA)
        write.table(r.mean, paste(rootOutputFilePrefix, "mean.tsv", sep=""), sep="\t", col.names=NA)
        write.table(r.sd, paste(rootOutputFilePrefix, "sd.tsv", sep=""), sep="\t", col.names=NA)
        write.table(r.cv, paste(rootOutputFilePrefix, "cv.tsv", sep=""), sep="\t", col.names=NA)
        write.table(r.sem, paste(rootOutputFilePrefix, "sem.tsv", sep=""), sep="\t", col.names=NA)
        write.table(r.pvals, paste(rootOutputFilePrefix, "p-value.tsv", sep=""), sep="\t", col.names=NA)
        write.table(r.plus, paste(rootOutputFilePrefix, "plus.tsv", sep=""), sep="\t", col.names=NA) # NA and "" are different, NA means not enough data, and M82=NA
        write.table(r.qvals, paste(rootOutputFilePrefix, "q-value.tsv", sep=""), sep="\t", col.names=NA)
        write.table(r.stars, paste(rootOutputFilePrefix, "stars.tsv", sep=""), sep="\t", col.names=NA) # NA and "" are different, NA means not enough data, and M82=NA

        ##################################
        # Create and write out a table of IL data, all traits.
        ##################################

        printLog("Making file (_summaryData.tsv) of p-values, means, sd, stars, and plus for all traits.\n")
        traits = colnames(r.pvals)
        df = NULL
        for (t in traits)
            {
            dfx = data.frame(trait=t, genotype=rownames(r.N), N=r.N[,t], mean=r.mean[,t], sd=r.sd[,t], cv=r.cv[,t],
                sem=r.sem[,t], pvalue=r.pvals[,t], qvalue=r.qvals[,t], stars=r.stars[,t], plus=r.plus[,t],
                stringsAsFactors=FALSE)
            df = rbind(df, dfx)
            }
        # Sort them nicely.
        ord = sortILs(genotype=df$genotype, trait=df$trait, first=list(genotype=c("M82", "PENN")),
            lead0=c("genotype", "trait"))
        df = df[ord,]
        rownames(df) = NULL
        write.table(df, file=paste(rootOutputFilePrefix, "summaryData.tsv", sep=""), sep="\t", row.names=FALSE)

        ##################################
        # Write the above table without the p-value, adj p-value, sd, stars, and plus columns.
        ##################################
        printLog("Making file (_means.tsv) of means for all traits, all genotypes.\n")
        df$trait = factor(df$trait)
        df$genotype = factor(df$genotype)
        df = df[, -match("pvalue", colnames(df))]
        df = df[, -match("qvalue", colnames(df))]
        df = df[, -match("sd", colnames(df))]
        df = df[, -match("stars", colnames(df))]
        df = df[, -match("plus", colnames(df))]
        colnames(df) = c("trait", "genotype", "mean")
        # Sort them nicely.
        ord = sortILs(genotype=df$genotype, trait=df$trait, first=list(genotype=c("M82", "PENN")),
            lead0=c("genotype", "trait"))
        df = df[ord,]
        rownames(df) = NULL
        write.table(df, file=paste(rootOutputFilePrefix, "means.tsv", sep=""), sep="\t", row.names=FALSE)

        ##################################
        # Create and write out a table of ILs significant in any trait.
        ##################################

        printLog("Making file (_sigGenotypes.tsv) containing list of all significant genotypes for each trait.\n")
        df = NULL
        for (t in traits)
            {
            sigIdxs = which(nchar(r.plus[,t]) > 0)
            if (length(sigIdxs) > 0)
                {
                dfx = data.frame(trait=t, genotype=rownames(r.N)[sigIdxs], N=r.N[sigIdxs,t],
                    mean=r.mean[sigIdxs,t], sd=r.sd[sigIdxs,t], cv=r.cv[sigIdxs,t],
                    sem=r.sem[sigIdxs,t], pvalue=r.pvals[sigIdxs,t], qvalue=r.qvals[sigIdxs,t],
                    stars=r.stars[sigIdxs,t], plus=r.plus[sigIdxs,t], stringsAsFactors=FALSE)
                df = rbind(df, dfx)
                }
            }
        # Sort them nicely.
        ord = sortILs(genotype=df$genotype, trait=df$trait, first=list(genotype=c("M82", "PENN")),
            lead0=c("genotype", "trait"))
        df = df[ord,]
        rownames(df) = NULL
        write.table(df, file=paste(rootOutputFilePrefix, "sigGenotypes.tsv", sep=""), sep="\t", row.names=FALSE)

        ##################################
        # Write the above table without the p-value, adj p-value, sd, stars, and plus columns.
        ##################################

        printLog("Making file (_sigMeans.tsv) of means for significant traits/genotypes.\n")
        df$trait = factor(df$trait)
        df$genotype = factor(df$genotype)
        df = df[, -match("pvalue", colnames(df))]
        df = df[, -match("qvalue", colnames(df))]
        df = df[, -match("sd", colnames(df))]
        df = df[, -match("stars", colnames(df))]
        df = df[, -match("plus", colnames(df))]
        colnames(df) = c("trait", "genotype", "mean")
        write.table(df, file=paste(rootOutputFilePrefix, "sigMeans.tsv", sep=""), sep="\t", row.names=FALSE)
        printLog("Finished analysis for analyzeDeriv = ", analyzeDeriv, " and rotMethod= ", rotMethod, "\n", sep="")
        }
    }

##################################
# Finished.
##################################
