#######################################################################################
# This file contains R code to plot results of model fitting to root phenotype data
# from the plate rotation experiments of the tomato M82/PENN introgression lines done
# by Edgar Spalding's lab.  Those were time course experiments that produced 34 time
# points over a course of two hours.
#
# This file runs a loop to process each of six input files containing model analysis
# output from each of the two raw input data files and for each of those, for the
# three different methods of selecting the data to be processed:
#   "THRU" - process data for roots rotated THROUGH the gravity vector
#   "AWAY" - process data for roots rotated AWAY from the gravity vector
#   "ALL" - process data for both rotations
# The model analysis output data file we read to obtain plotting info has a
# filename of the form: *_<rawDataFileBasename>_<rotMethod>_min4_inputDataFiltered.tsv
#
# Output filenames are different for each analysis results input file.
#
# To run this code, select all the code in the file and execute it.  Check the
# console for errors.
#
# Revision history:
# Author    Date        Description
# --------  -------     ---------------------------------------------------------------
# Ted Toal  11Mar15     Create this file by splitting this code out from the analysis
#                       code file Analysis_Rot.R.
# Ted Toal  20Apr2015   Add rotMethod and loop.
#######################################################################################

# Include the root R source code file.
R_INCLUDE_DIR = Sys.getenv("R_INCLUDE_DIR")
if (R_INCLUDE_DIR == "")
    stop("Set environment variable R_INCLUDE_DIR or assign that variable to a path here.")
source(paste(R_INCLUDE_DIR, "Include_RootRsourceFile.R", sep=""))

# Source our basic utility functions file.
source(Include_UtilityFunctions)

# Load libraries we need.
addThisLibrary("xlsx")
addThisLibrary("cluster")

# Source other files we need.
source(Include_TextFunctions)
source(Include_Heatmaps)
source(Include_xlsx)
source(Include_TomatoILsBILs)
source(Include_phenoRNAseqRootSep2011)
source(Include_phenoAnalysis)

# Set working directory.
setwd(SpaldingAnalysisDir)

# Set this TRUE to force all tip angles of each genotype to be altered by subtracting
# the angle at time 0, giving all genotypes a tip angle of 0 at time 0.  This is so
# that a reasonable tip angle comparison can be done.  In this case, trait T0 is
# removed.
alignAnglesAtT0 = TRUE

# SET THIS TO THE DATE OF THE ANALYSIS WHOSE DATA IS TO BE PLOTTED.  This date
# prefixes all input and all output filenames AND IS THE OUTPUT DIRECTORY NAME.
# If you don't want to overwrite plot files already generated, you should save
# them before running this.
fileDate = "21May2015"

# SET THIS TO A NAME TO FOLLOW fileDate IN THE I/O FOLDER NAME TO IDENTIFY
# WHAT DISTINGUISHES THIS RUN FROM OTHERS.
{
if (!alignAnglesAtT0)
    folderSfx = "_AwayThru"
else
    folderSfx = "_AlignAtT0"
}

# Input/output folder name.
ioFolder = paste(fileDate, folderSfx, sep="")

# P-value breakpoints for increasing numbers of stars.
pValueBreaks = c(0.001, 0.01, 0.05, 99)
theStars = c("***", "**", "*", "")
# Significant p-values.
sigp = rev(pValueBreaks)[-1]

# Analysis and plotting option settings.  Change these as desired.
paper = FALSE # Use TRUE for image for paper, FALSE for ourselves.
addNumSampleAnnotation = !paper # Use FALSE to suppress annotation.
blankInsignificantNames = FALSE # Use TRUE to blank the names of non-sig. ILs on X-axis label.
colorBarsBySignificance = TRUE # Use TRUE to color bars darker if they are more significant
colorTheseGenotypeBars = c(parents) # To color all RNA-seq'd bars use c("M82", "PENN", "2-2", "5-2", "7-4-1", "7-5", "7-5-5", "8-2") # Color these bars special colors.
colorOfThoseBars = c(parentColors) # To color all RNA-seq'd bars use c("red", "green", "blue", "blue", "blue", "blue", "blue", "blue") # The colors to use.
addErrorBars = TRUE # Use FALSE to suppress error bars.
showPlusForUncorrectedSigPvalues = TRUE # Use FALSE to suppress "+" signs
makeHeatmapSubsetGTs = list()

# Number of standard deviations beyond which to call a data point an outlier.
xsd = 3.0

################################################################################
################################################################################
################################################################################
# Loop for each rotation method:
for (rotMethod in c("ALL", "THRU", "AWAY"))
    {
    # Set criteria for data filtering.
    {
    if (rotMethod == "ALL")
        minRootsPerIL = 4
    else
        minRootsPerIL = 3
    criteriaCode = paste("min", minRootsPerIL, sep="")
    }

    # Loop for both input files (tip angles and derivatives).
    for (analyzeDeriv in c(FALSE, TRUE))
        {
        # Set filename prefix, which prefixes both input and output files, and includes
        # the input/output folder at the beginning of it.
        {
        if (analyzeDeriv)
            rootIOprefix = "allILderiv"
        else
            rootIOprefix = "allILtipangles"
        rootIOprefix = paste(fileDate, "_", rootIOprefix, "_", rotMethod, "_", criteriaCode, "_", sep="")
        rootIOprefix = paste(ioFolder, rootIOprefix, sep=PATHSEP)
        }

        # Input data filenames.
        traitsFile = paste(rootIOprefix, "traits.tsv", sep="")
        groupsFile = paste(rootIOprefix, "traitGroups.tsv", sep="")
        vettedInputDataFile = paste(rootIOprefix, "inputDataFiltered.tsv", sep="")
        modelSummaryDataFile = paste(rootIOprefix, "summaryData.tsv", sep="")

        # Get trait names.

        dfTraits = read.table(traitsFile, sep="\t", header=TRUE, row.names=1, stringsAsFactors=FALSE)
        traits = rownames(dfTraits)
        ntraits = length(traits)
        cat("There are", ntraits, "traits in total: ", traits, "\n")

        # Get groups.

        dfGroups = read.table(groupsFile, sep="\t", header=TRUE, row.names=1, stringsAsFactors=FALSE)

        # Read the filtered input data file.  This file has already been thoroughly
        # vetted for correctness by an earlier pre-processing stage, and was also
        # filtered by the modelling code to remove data that was insufficient for
        # modelling.

        data.all = read.table(vettedInputDataFile, sep="\t", header=TRUE, row.names=NULL, stringsAsFactors=FALSE)

        # Read the model output giving significance of different traits in different ILs.

        dfModelResults = read.table(modelSummaryDataFile, sep="\t", header=TRUE, row.names=NULL, stringsAsFactors=FALSE)

        # Get genotypes.

        genotypes = unique(data.all$genotype)
        ngenotypes = length(genotypes)
        cat("There are", ngenotypes, "genotypes in total:\n")
        cat(genotypes, "\n", fill=70, labels="   ")

        ##################################
        ## Make some plots to allow us to look for bad data.

        pdf(paste(rootIOprefix, "dataplots.pdf", sep=""), height=10, width=10)
        for (t in traits)
            {
            cat("Data plot for trait", t, "\n")

            # Get filtered data frame and change name of the current trait column to "trait"
            data.r = data.all[, c("genotype", t)]
            colnames(data.r)[2] = "trait"

            # X will be the genotype, Y will be the data value or mean.  Shape of the point will
            # indicate mean vs data point, round for means, triangle for data points.  Use black
            # for IL mean.  Plot only those data points that are more than 2 sd's away from IL mean.
            xlim = c(1, ngenotypes)
            ylim = range(pretty(c(0, 1.2*range(data.r$trait))))
            xmax = xlim[2]
            ymax = ylim[2]
            ydel = 0.03*ymax
            ypos = ymax
            meanPerIL = tapply(data.r$trait, data.r$genotype, mean)
            # Report X values by meanPerIL, apply this sorting order to remaining points() calls.
            xposIL = order(meanPerIL)
            names(xposIL) = genotypes[xposIL]
            plot(NA, type="n", xlim=xlim, ylim=ylim, xaxt="n",
                main=paste("Data for trait", t),
                xlab="", ylab=dfTraits[t, "units"])
            axis(1, at=1:length(xposIL), labels=names(xposIL), font=2, las=3, cex.axis=0.8) 
            pchDot = 20
            points(1:length(xposIL), meanPerIL[names(xposIL)], pch=pchDot, col="black")
            lines(1:length(xposIL), meanPerIL[names(xposIL)], col="black")
            legend(1, ypos, "IL Means", bty="n", pch=pchDot, col="black")
            ypos = ypos - ydel
            meanPerIL = tapply(data.r$trait, data.r$genotype, mean)
            points(1:length(xposIL), meanPerIL[names(xposIL)], pch=pchDot)
            lines(1:length(xposIL), meanPerIL[names(xposIL)])
            ypos = ypos - ydel
            sdPerIL = tapply(data.r$trait, data.r$genotype, sd)
            moreThanXSD = (abs(data.r$trait - meanPerIL[data.r$genotype]) > xsd*sdPerIL[data.r$genotype])
            x = data.r$genotype[moreThanXSD]
            y = data.r$trait[moreThanXSD]
            pchTriangle = 17
            points(xposIL[x], y, pch=pchTriangle)
            legend(1, ypos, paste("data pt >", xsd, "SDs"), bty="n", pch=pchTriangle)
            ypos = ypos - ydel
            }
        dev.off()

        ##################################
        ## Bar plots.

        # Loop for each trait and make a bar plot for each one.
        pdf(paste(rootIOprefix, "barplots.pdf", sep=""), width=11, height=10)
        for (t in traits)
            {
            cat("Bar chart for trait", t, "\n")

            # Get filtered data frame and change name of the current trait column to "trait"
            data.r = data.all[, c("genotype", t)]
            colnames(data.r)[2] = "trait"

            # Get bar and star data and sort it.  Remove NaN values.
            dfResults = dfModelResults[dfModelResults$trait == t,]
            rownames(dfResults) = dfResults$genotype
            dfResults = dfResults[genotypes,]
            mean_values = dfResults$mean
            names(mean_values) = dfResults$genotype
            err_values = dfResults$sem
            names(err_values) = dfResults$genotype
            bar_desc = "Bars = std. err. (sigma) from lm/glht"
            stars = dfResults$stars
            names(stars) = dfResults$genotype
            stars[is.na(stars)] = ""
            plus = dfResults$plus
            names(plus) = dfResults$genotype
            plus[is.na(plus)] = ""

            # Sort by mean value of genotype.
            ord = order(mean_values)
            mean_values = mean_values[ord]
            err_values = err_values[ord]
            stars = stars[ord]
            plus = plus[ord]

            # Assign colors for bars of each different significance category.
            mean_values = dfResults$mean
            names(mean_values) = dfResults$genotype
            # Each of these colors also comes in a range of tints from light to dark, which are
            # obtained by adding a digit 1(light)..4(dark) to the end of the name. 
            sigBarColors = c("gray70", "gray55", "gray40", "thistle4", "darkolivegreen", "dodgerblue", "sienna3")
            sigBarColors = sigBarColors[1:length(sigp)] # We only need this many colors.

            # Set bar colors.
            barcolors = rep("gray90", length(mean_values))
            names(barcolors) = names(mean_values)
            num_stars = nchar(stars)
            if (colorBarsBySignificance)
                barcolors[names(num_stars)[num_stars > 0]] = sigBarColors[num_stars[num_stars > 0]]
            if (length(colorTheseGenotypeBars) > 0)
                barcolors[colorTheseGenotypeBars] = colorOfThoseBars

            # Optionally use blank names for x-axis labels of ILs that aren't significant, except
            # M82 and PENN are always shown.
            barnames = names(mean_values)
            if (blankInsignificantNames)
                barnames[stars[barnames] == "" & barnames != "M82" & barnames != "PENN"] = ""

            # Make some text size magnification estimates.
            cex = 50/length(mean_values) # A weird guess...
            dx = 0
            # Here's a correction for when there are very few bars.
            if (cex > 1)
                {
                cex = 1
                dx = -4/length(mean_values)
                }

            # Make a bar plot.
            ylim = range(pretty(c(0, 1.4*range(mean_values))))
            ymax = ylim[2]
            ydel = diff(ylim)/25
            ypos = ymax
            yvals = mean_values
            xvals = 1:length(yvals)
            xdel = 3
            barplot(yvals, las=3, cex.names=cex, font.axis=2, col=barcolors, space=0, ylim=ylim,
                names.arg=barnames,
                main=paste("\n", dfTraits[t, "desc"]),
                xlab="", ylab=dfTraits[t, "units"])

            # Show data filtering criteria.
            if (!paper)
                {
                dataFilterDesc = paste("Data filtering: >=", minRootsPerIL, "roots/IL")
                legend(length(xvals)/3, ymax-ydel, dataFilterDesc, cex=0.8, bty="n", text.col="black", yjust=0.7)
                }
        
            # Add error bars.
            if (addErrorBars)
                {
                arrows(xvals-0.5, yvals+err_values, xvals-0.5, yvals-err_values, angle=90, code=3, length=0.025)
                if (!paper)
                    legend(length(xvals)/3, ymax-2*ydel, bar_desc, cex=0.8, bty="n", text.col="black", yjust=0.7)
                }

            # Add stars above bars.
            for (x in xvals)
                {
                il = names(mean_values)[x]
                if (stars[il] != "")
                    {
                    y = mean_values[il] + (ydel/3)*sign(mean_values[il])
                    color = sigBarColors[num_stars[il]]
                    if (addErrorBars)
                        y = y + err_values[il]*sign(mean_values[il])
                    text(x+dx, y, stars[il], cex=2*cex, font=2, srt=90, pos=4, offset=0, col=color)
                    }
                }

            # Add plusses above bars for Siobhan results.
            if (showPlusForUncorrectedSigPvalues)
                {
                for (x in xvals)
                    {
                    il = names(mean_values)[x]
                    if (plus[il] != "")
                        {
                        y = mean_values[il] + ydel*sign(mean_values[il])
                        if (addErrorBars)
                            y = y + err_values[il]*sign(mean_values[il])
                        text(x-.5, y, plus[il], cex=cex, font=2, srt=90, pos=4, offset=0, col="blue")
                        }
                    }
                }

            # Add a stars/plusses legend.
            for (i in 1:length(sigp))
                {
                if (showPlusForUncorrectedSigPvalues)
                    {
                    plusses = paste(rep("+", i), collapse="")
                    legend(xdel, ypos, plusses, cex=1, bty="n", text.col=sigBarColors[i],
                        xjust=1, yjust=0.74)
                    }
                stars = paste(rep("*", i), collapse="")
                legend(0, ypos-ydel/4, stars, cex=1.5, bty="n", text.col=sigBarColors[i], yjust=0.7)
                legend(length(xvals)/15, ypos, paste("p <", sigp[i]), cex=1, bty="n", text.col=sigBarColors[i],
                    yjust=0.7)
                ypos = ypos - ydel
                }

            # Add annotation of number of samples to bars.
            if (addNumSampleAnnotation)
                {
                numRoots = tapply(1:nrow(data.all), data.all$genotype, length)
                numRoots[is.na(numRoots)] = 0
                for (x in xvals)
                    {
                    il = names(mean_values)[x]
                    text(x-0.5, 1, numRoots[il], cex=0.5, font=2, srt=90, pos=4, offset=0, col="black")
                    }
                }
            }
        dev.off()

        ##################################
        ### Heat maps and dendrograms

        cat("\nMaking heat map\n")

        # If this is an AWAY or THRU heat map, read the model data for the opposite
        # one also.  We want the AWAY and THRU heat maps to contain the same ILs.
        # To do that we must add NA rows to this heat map for ILs not present in
        # this data set but significant in the opposite one.
        dfModelResults.other = NULL
        if (rotMethod != "ALL")
            {
            rotS = paste("_", rotMethod, "_", sep="")
            rotS.other = paste("_", ifelse(rotMethod == "THRU", "AWAY", "THRU"), "_", sep="")
            modelSummaryDataFile.other = sub(rotS, rotS.other, modelSummaryDataFile)
            dfModelResults.other = read.table(modelSummaryDataFile.other, sep="\t",
                header=TRUE, row.names=NULL, stringsAsFactors=FALSE)
            }

        # Get a list of IL's that have at least one star in at least one trait
        # of traits.
        sigILs = c()
        for (t in traits)
            {
            dfResults = dfModelResults[dfModelResults$trait == t,]
            rownames(dfResults) = dfResults$genotype
            stars = dfResults$stars
            names(stars) = dfResults$genotype
            stars[is.na(stars)] = ""
            sigILs = union(sigILs, names(stars)[stars != ""])

            # For AWAY and THRU, get significant ILs for the other data set too
            # and add those to sigILs.
            if (rotMethod != "ALL")
                {
                rotS = tolower(paste(rotMethod, "_", sep=""))
                rotS.other = tolower(paste(ifelse(rotMethod == "THRU", "AWAY", "THRU"), "_", sep=""))
                t.other = sub(rotS, rotS.other, t)
                dfResults.other = dfModelResults.other[dfModelResults.other$trait == t.other,]
                rownames(dfResults.other) = dfResults.other$genotype
                stars = dfResults.other$stars
                names(stars) = dfResults.other$genotype
                stars[is.na(stars)] = ""
                sigILs = union(sigILs, names(stars)[stars != ""])
                }
            }

        # Create a data frame of number of stars for each IL/trait combination in sigILs and
        # traits vectors.  For each one, test the trait's mean value in that IL with the mean
        # value for M82, and use a NEGATIVE NUMBER OF STARS if the trait value is less than
        # M82.  ILs for which the qvalue field is NA have nstars set to NA, which causes a
        # gray heatmap square below.
        dfStars = as.data.frame(matrix(NA, nrow=length(sigILs), ncol=length(traits),
            dimnames=list(sigILs, traits)))
        for (t in traits)
            {
            dfResults = dfModelResults[dfModelResults$trait == t,]
            rownames(dfResults) = dfResults$genotype
            dfResults = dfResults[sigILs,]
            stars = dfResults$stars
            names(stars) = dfResults$genotype
            nstars = nchar(stars)
            means = dfResults$mean
            M82mean = dfModelResults$mean[dfModelResults$trait == t & dfModelResults$genotype == "M82"]
            lessThanM82 = (means < M82mean)
            nstars = nstars * ifelse(lessThanM82, -1, +1)
            nstars[is.na(dfResults$qvalue)] = NA
            dfStars[sigILs, t] = nstars
            }

        # Sort data frame ILs in order.
        ord = sortILs(rownames(dfStars))
        dfStars = dfStars[ord,]

        # Write the data to a file.
        write.table(dfStars, file=paste(rootIOprefix, "Heatmap.tsv", sep=""), sep="\t", row.names=TRUE)

        # Make the column names shorter so the heatmap labels aren't obtrusive.
        colnames(dfStars) = sub("^[^_]*_", "", colnames(dfStars))

        # Make colors to use on the heatmap.  Make multiple copies of the heatmap, with different
        # colors.
        colorSets = list(
            list(ltM82color="green4", gtM82color=parentColors["M82"], NOTSIGcolor="black", NAcolor="darkslategray",
                 ltM82intensity=c(1,15,30), gtM82intensity=c(205,230,255),
                 ltM82colorName=parentColorNames["M82"], gtM82colorName=parentColorNames["PENN"],
                 groupRectColor="blue",
                 subtitle="(red/green version for the non-color-impaired)",
                 filenameSuffix="_NonColorImpaired"),
        #    list(ltM82color="#D55E00", gtM82color="#009E73", NOTSIGcolor="black", NAcolor="darkslategray",
            list(ltM82color="#F56500", gtM82color="#00A275", NOTSIGcolor="black", NAcolor="darkslategray",
                 ltM82intensity=c(1,25,50), gtM82intensity=c(190,222,255),
                 ltM82colorName="red", gtM82colorName="blugrn",
                 groupRectColor="orange",
                 subtitle="(red/blue version for the color-impaired)",
                 filenameSuffix="_ForColorImpaired1"),
        #    list(ltM82color="#F0E442", gtM82color="#0072B2", NOTSIGcolor="black", NAcolor="darkslategray", # "#009E73",
            list(ltM82color="#F0E442", gtM82color="#56B4E9", NOTSIGcolor="black", NAcolor="darkslategray", # "#009E73",
                 ltM82intensity=c(1,15,35), gtM82intensity=c(190,230,255),
                 ltM82colorName="yellow", gtM82colorName="blue",
                 groupRectColor="tan3",
                 subtitle="(yellow/blue version for the color-impaired)",
                 filenameSuffix="_ForColorImpaired2")
            )

        for (colorSet in colorSets)
            {
            cat("Making heatmap of significant ILs/traits:", colorSet$subtitle, "\n")

            # Put dendrograms and heatmap into a .pdf file.
            heatmapPdf = paste(rootIOprefix, "Heatmaps", colorSet$filenameSuffix, ".pdf", sep="")
            pdf(heatmapPdf, width=12, height=10)
    
            # Make colors to use on the heatmap.
            heatmapColors = makeHeatmapColors("#808080", colorSet$gtM82color, 127, colorSet$ltM82color, 127)
            xlabKey = paste("p-value threshold, ", colorSet$ltM82colorName, ": mean < M82 mean, ", colorSet$gtM82colorName, ": mean > M82 mean", sep="")
            nStarsBreaks = nchar(theStars)
            pvals = rep(NA, length(theStars))
            pvals[nStarsBreaks+1] = sub("0+$", "", as.character(pValueBreaks))
            pvals[1] = "1" # Change 99 to 1
            #pvals
            xvalKey.nstars = c(NA, 3, 2, 1, 0, 1, 2, 3)
            xvalKey = as.character(pvals[xvalKey.nstars+1])
            xvalKey[1] = "NA"
            #xvalKey
            # The first color is for value -4 = NA; following this are the values for
            # -3, -2, -1 stars (< M82), then 0 stars, then 1, 2, 3 stars (> M82).
            colors = c(colorSet$NAcolor, heatmapColors[colorSet$ltM82intensity],
                colorSet$NOTSIGcolor, heatmapColors[colorSet$gtM82intensity])   
            breaks = seq(-4, +3, length.out=length(colors)+1)
    
            # Make a plot.Heatmap callback function to plot rectangles around the trait groups.
            plot.Heatmap = function(xlim, ylim, nr, nc)
                {
                xRange = t(sapply(dfGroups$formula, function(S) range(as.integer(unlist(strsplit(gsub("T", "", S), "|", fixed=TRUE))))))
                rownames(xRange) = rownames(dfGroups)
                colnames(xRange) = c("xbot", "xtop")
                ybot = ylim[1] + abs(mean(c(ylim[1], 0))/2)
                ytop = mean(c(nr, ylim[2]))
                for (i in 1:nrow(xRange))
                    {
                    rect(xRange[i, "xbot"], ybot, xRange[i, "xtop"]+1, ytop, border=colorSet$groupRectColor, lwd=5)
                    text(mean(xRange[i,]), ybot, rownames(xRange)[i], pos=3, cex=1.3)
                    }
                }
            
            # Make a heatmap with no dendrograms.
            L = makeHeatmap(dfStars, colDendro=NULL, rowDendro=NULL, xlabKey=xlabKey, cexLabKey=1.05, cex.axisKey=1.45,
                xvalKey=xvalKey, margins=c(bottom=5,left=5,top=10,right=5),
                heatmapAreaSize=c(80,75), heatmapSize=c(55,78), keySize=c(60,35), keySpace=c(13,35), key=3,
                lftLabSize=c(15,0.5), btmLabSize=c(10,0.5), topLabSize=c(10,0.5),
                colsep=2:ncol(dfStars), rowsep=2:nrow(dfStars), sepcolor="gray", sepwidth=c(0.13,0.066),
                density.info="none", cexRow=1.2, cexCol=1.3, cex.main=1.3,
                valueForNA=-4, heatmapColors=colors, breaks=breaks,
                topLabCol=TRUE, btmLabCol=TRUE, lftLabRow=TRUE, rgtLabRow=FALSE, srtLabCol=90,
                titleRowDendrogram="ILs clustered by abs. Pearson corr. of each trait's p-value for comparison to M82",
                titleHeatmap="Heatmap: ILs with significant p-value in trait comparison to M82\nILs clustered by abs. Pearson corr. of p-value",
                subtitleHeatmap=colorSet$subtitle, mainKey="", line.xlabKey=3, plot.Heatmap=plot.Heatmap)

            # Make a heatmap with dendrograms.
            makeHeatmap(dfStars, colDendro=NULL, rowDendro=NA,
                xlabKey=xlabKey, xvalKey=xvalKey, cexLabKey=0.8, cex.axisKey=1.2,
                margins=c(bottom=5,left=5,top=10,right=5),
                heatmapAreaSize=c(60,70), heatmapSize=c(70,70), keySize=c(100,15), keySpace=c(0,45), key=3,
                lftLabSize=c(10,0.5), rgtLabSize=c(0,0), btmLabSize=c(10,0.5), topLabSize=c(0,0.5),
                colsep=2:ncol(dfStars), rowsep=2:nrow(dfStars), sepcolor="gray", sepwidth=c(0.13,0.066),
                density.info="none", cexRow=1.0, cexCol=1.0, cex.main=1.2,
                valueForNA=-4, heatmapColors=colors, breaks=breaks, rowDendroSize=c(90,10),
                topLabCol=FALSE, btmLabRow=TRUE, lftLabRow=TRUE, rgtLabRow=FALSE, srtLabCol=90,
                titleRowDendrogram="ILs clustered by abs. Pearson corr. of each trait's p-value for comparison to M82",
                titleHeatmap="Heatmap: ILs with significant p-value in trait comparison to M82\nILs clustered by abs. Pearson corr. of p-value",
                subtitleHeatmap=colorSet$subtitle, mainKey="", line.xlabKey=3)
    
            # Make subset heatmaps.
            for (subsetGTs in makeHeatmapSubsetGTs)
                {
                dfSubset = dfStars[subsetGTs,]
                makeHeatmap(dfSubset, colDendro=NULL, rowDendro=NULL,
                    xlabKey=xlabKey, xvalKey=xvalKey, cexLabKey=0.8, cex.axisKey=1.2,
                    margins=c(bottom=50,left=30,top=10,right=35),
                    heatmapAreaSize=c(35,30), heatmapSize=c(50,20), keySize=c(90,20), keySpace=c(10,30), key=3,
                    colsep=2:ncol(dfSubset), rowsep=2:nrow(dfSubset), sepcolor="gray", sepwidth=c(0.13,0.066),
                    density.info="none", cexRow=1.0, cexCol=1.0, cex.main=1.2,
                    valueForNA=-4, heatmapColors=colors, breaks=breaks,
                    topLabCol=TRUE, btmLabCol=FALSE,
                    lftLabRow=TRUE, rgtLabRow=FALSE, lftLabSize=c(30,0.5), srtLabCol=90,
                    titleHeatmap="Heatmap: subset of ILs with significant p-value in trait comparison to M82",
                    subtitleHeatmap=colorSet$subtitle, mainKey="", line.xlabKey=3)
                }
    
            # Close the pdf file.
            dev.off()
            }
        }
    }

##################################
# Finished.
