#######################################################################################
# This file contains R definitions and functions relating to text printing and
# manipulation.
# Author: Ted Toal
# Lab: Brady
#######################################################################################

cat("Including Include_TextFunctions.R\n")

#######################################################################################
# Suppress specific messages (regular or warning) coming from a function call.
#
# Arguments:
#   expr: an R expression to evaluate.
#   ...: one or more regular expressions to be applied to any message generated as a
#       result of evaluating "expr".  If any expression matches the message, it is
#       suppressed, else it is printed.  Note that messages sometimes end in "\n" so
#       make sure that is matched if the regular expression ends with $.  Also note
#       that multiple regular expressions can be specified as either separate arguments
#       or as single arguments that are vectors of regular expressions.
#   warningsToErrors: if TRUE, warnings not matched by the regular expressions are
#       turned into errors.
#
# Returns: whatever is returned by evaluation of the expression.
#
# Notes: look at implementation of function message() and read the R help on the
# functions used in it.  Then search on web, e.g.
# http://adv-r.had.co.nz/beyond-exception-handling.html
#
# Example:
#     t.cor = muffleMessages(
#            {
#            cor.test(gene1.log2Counts, gene2.log2Counts, alternative="two.sided", method=corMethod)
#            }, "Cannot compute exact p-value with ties", warningsToErrors=TRUE)
#######################################################################################
muffleMessages = function(expr, ..., warningsToErrors=FALSE)
    {
    REs = list(...)

    # Search for a regular expression matching msg and return TRUE iff found.
    anyMatch = function(msg)
        {
        for (REvec in REs)
            for (RE in REvec)
                if (grepl(RE, msg))
                    return(TRUE)
        return(FALSE)
        }

    # Search for a restart and invoke it if found.
    doRestart = function(restart)
        {
        r = findRestart(restart) 
        if (is.null(r))
            return()
        invokeRestart(r)
        }

    # Establish simpleMessage and simpleWarning handlers, then evaluate the expression.
    withCallingHandlers(
        simpleMessage=function(m) { if (anyMatch(conditionMessage(m))) doRestart("muffleMessage") },
        simpleWarning=function(w)
            {
            if (anyMatch(conditionMessage(w)))
                doRestart("muffleWarning")
            if (warningsToErrors)
                stop(paste("  (Warning in ", Reduce(paste, deparse(w$call)), ")   : ", w$message, sep=""), call.=FALSE)
            },
        expr)
    }
       
#######################################################################################
# Call cat() with the arguments, followed by flush.console() to ensure immediate output
# of the data.  Use this instead of cat() when long computations occur and you want to
# print something just before doing the computation.
#######################################################################################
catnow = function(...)
	{
	cat(...)
	flush.console()
	return(invisible(0))
	}

#######################################################################################
# Set the name of an output log file used by the printLog() function below.  If you
# don't set it, the default log file name is "defaultLogFile.txt".
#######################################################################################
outLogFile = "defaultLogFile.txt"
setDefaultLogFile = function(filename)
	{
	assignGlobal("outLogFile", filename)
	}

#######################################################################################
# Write the arguments to BOTH the console and the log file.  Arguments are like the
# cat() function arguments.  Use "file=filename" to specify a particular log file,
# otherwise the default log file set above is used.  You can also write the same output
# to multiple log files by using a vector of filenames for "file", and using NA in that
# vector causes output to the default log file set above.  To create a new log file and
# delete the old one, make the first call to this function use "append=FALSE".
# Note: output to the console can be suppressed by specifying toConsole=FALSE.  Also,
# if you set file="" (empty string), nothing is written to the log file.
#######################################################################################
printLog = function(..., sep=" ", append=TRUE, file=NA, toConsole=TRUE)
	{
	file[is.na(file)] = outLogFile
	if (toConsole)
	    {
	    cat(..., sep=sep)
	    flush.console()
	    }
	for (aFile in file)
	    if (aFile != "")
    	    cat(..., sep=sep, file=aFile, append=append)
	}

#######################################################################################
# Use this handy function to print out long strings of text, indented by some amount
# with the text reflowed to fit in a smaller space.
#
# Concatenate all unnamed arguments together into a single long string separated by the
# string given by sep, then break up that string into multiple lines, each prefixed by
# the string given by prefix, and each of length no more than len, breaking lines at
# spaces or \n, or if absolutely necessary, at an arbitrary character.  Return the
# resulting strings as a single long string containing \n at the end of each line
# including the last.
#######################################################################################
fillText = function(..., sep=" ", prefix="", len=80)
    {
    strs = unlist(strsplit(paste(..., sep=sep), "\n", fixed=TRUE))
	result = ""
	if (nchar(prefix) >= len || len < 1)
		stop("Invalid prefix or len in fillText()")
	# Loop for each line that is separated by "\n".
    for (str in strs)
    	{
    	line = prefix
    	tokens = unlist(strsplit(str, " ", fixed=TRUE))
    	for (token in tokens)
    		{
    		# Loop constant: there is room for at least one character before reaching len characters
    		# in line.

    		# If line has more than prefix on it, append a space to it before adding token.
    		if (line != prefix)
    			{
    			# If appending a space makes the line length be len, start a new line.
    			if (nchar(line) + 1 < len)
    				line = paste(line, " ", sep="")
    			else
    				{
    				result = paste(result, line, "\n", sep="")
    				line = prefix
    				}
				}
			# Loop outputting token characters to lines, until token is empty.
    		while (token != "")
    			{
				# If token fits leaving at least one character, add it to line.
				if (nchar(line) + nchar(token) + 1 <= len)
					{
					line = paste(line, token, sep="")
					token = ""
					}
				# Else if line is not empty, start a new line.
				else if (line != prefix)
					{
					result = paste(result, line, "\n", sep="")
					line = prefix
					}
				# Else line is currently empty, add as many chars of token as possible to line and
				# start a new line.
				else
					{
					roomFor = len - nchar(line)
					line = paste(line, substr(token, 1, roomFor), sep="")
					token = substr(token, roomFor+1, 1000000L)
					result = paste(result, line, "\n", sep="")
					line = prefix
					}
				}
			}
		# If line is not empty, append it to the results.
		if (line != prefix)
			result = paste(result, line, "\n", sep="")
    	}
    return(result)
    }

#######################################################################################
# Print a small descriptive string followed by a possible-long descriptive string that
# may need to be broken up among multiple lines.  Use this for printing (to the console)
# things like:
#   AT7G01929     blah-blah-blah a long long description
# Here, "AT7G01929" is string1 and the rest is string2.
#
# Print string2 prefixed by string1.  Print string2 starting on the line AFTER string1,
# and fill those lines by using fillText to print string2, with the specified prefix
# string on each line.  If, however, string2 is NA or empty, just print "string1: (NONE)"
# if showNONE is TRUE.  If string2 consists of more than one string, they are collapsed
# into one string, using the specified collapse string as a separator.
#######################################################################################
catDescriptiveString = function(string1, string2, prefix, collapse=", ", showNONE=FALSE)
    {
    string2[is.na(string2) | string2 == "NA"] = ""
    string2 = paste(string2, collapse=collapse)
    if (gsub(collapse, "", string2) == "")
        {
        if (showNONE)
            cat(string1, " (NONE)\n")
        }
    else
        {
        cat(string1, "\n")
        cat(fillText(string2, prefix=prefix))
        }
    }

#######################################################################################
# Vector paste function.  Paste together elements of v, optionally ignoring empty
# strings and optionally pasting only one copy of duplicate strings.
#######################################################################################
vpaste = function(v, sep="", ignoreEmpty=TRUE, oneOfDuplicates=TRUE)
    {
    if (ignoreEmpty)
    	v = v[v != ""]
    if (oneOfDuplicates)
    	v = unique(v)
    return(paste(v, sep="", collapse=sep))
    }

#######################################################################################
# Expression paster.  This lets you make strings for plot labels containing formatted
# text such as subscripts, bold, math notation, etc.  See ?plotmath.  A limitation of
# the functionality described at ?plotmath is easily seen with this example:
#       x = 3.4
#       plot(1:10, main=expression(x^2))
# The plot title contains "x" with a 2 superscript.  What if you want 3.4 with a 2
# superscript?  This function lets you do it, as follows:
#       x = 3.4
#       plot(1:10, main=epaste(x, "^2"))
# To get the functionality where the title is "x" with a 2 superscript, do this:
#       plot(1:10, main=epaste("x^2"))
# If you wanted the title "x^2", do this:
#       plot(1:10, main=epaste("'x^2'"))
# And if you wanted the title "3.4^2", do this:
#       plot(1:10, main=epaste("'", x, "^2'"))
# Often you will need to surround text and characters you want to appear literally
# with quotes (use single since you will be using double quotes in the args to this
# function).  Also, the arguments to this function must form an expression that can
# be parsed, after pasting.  To do that, sometimes you need to include "*" in the
# expression to indicate pasting of expression elements, as in the following example
# (see ?plotmath).  Set the argument "debug" = TRUE to display the intermediate
# string to be parsed as an expression.
# Suppose you want the title "(x=3.4, y=5.6)" with superscript n on each number:
#       x = 3.4
#       y = 5.6
#       n = 3      
#       plot(1:10, main=epaste("('x=", x, "'^", n, "*', y=", y, "'^", n, ")"))
#######################################################################################
epaste = function(..., debug=FALSE)
    {
    L = list(..., sep="")
    S = do.call(paste, L)
    if (debug)
        cat("S='", S, "'\n", sep="")
    return(parse(text=S))
    }

#######################################################################################
# Combine multiple separated string lists into a single such list.  The elements of
# each list are split apart, unique elements are chosen and sorted, then they are
# re-pasted with a separator.
#
# Arguments:
#   elements: vector of strings, each string being a collection of elements separated
#       by "sep".
#   sep: character or string that separates each element of each string in "elements".
#       This same character or string is used to rejoin the strings.  Must not be a
#       regular expression.
#   sortOrder: a function to provide sorting order for the elements.  If NULL, no
#       sorting is done.  Default is order(), providing standard sorting order.
#
# Returns: a single string that is a collection of elements separated by "sep".
#######################################################################################
combineElements = function(elements, sep=",", sortOrder=order)
    {
    elements = unique(unlist(strsplit(elements, sep, fixed=TRUE), use.names=FALSE))
    if (!is.null(sortOrder))
        elements = elements[sortOrder(elements)]
    return(paste(elements, collapse=sep))
    }

#######################################################################################
# Read a FASTA file and extract sequence for each ID, reading nLinesAtOnce lines at a
# time and processing them.
# Return a vector of sequences, with the name being the sequence ID.
#######################################################################################
readFasta = function(filename, nLinesAtOnce=10000)
    {
    seqs = c()
    seqFile = file(filename, "r") 
    txt = readLines(seqFile, nLinesAtOnce)
    while (length(txt) > 0 && substring(txt[1], 1, 1) == ">")
        {
        # Find all sequence start lines.
        idLines = which(grepl("^>", txt))
        # If there are at least nLinesAtOnce of text, save text for last sequence,
        # which may be incomplete, and remove it from txt vector.
        lastSeqTxt = c()
        if (length(txt) >= nLinesAtOnce)
            {
            lastSeq = idLines[length(idLines)]
            idLines = idLines[-length(idLines)]
            lastSeqTxt = txt[lastSeq:length(txt)]
            txt = txt[1:(lastSeq-1)]
            }
        # Get rid of "*" stop codon.
        txt = sub("*", "", txt, fixed=TRUE)
        # Get sequence IDs.
        seqIDs = sub("^>([^ \t\r]+)[ \t]?.*$", "\\1", txt[idLines])
        # Concatenate sequence lines for each sequence ID.
        seqFirstTxt = idLines+1
        seqLastTxt = c(idLines[-1]-1, length(txt))
        seqTxt = sapply(1:length(seqFirstTxt), function(i)
            paste(txt[seqFirstTxt[i]:seqLastTxt[i]], collapse=""))
        names(seqTxt) = seqIDs
        # Append these sequences to the seqs vector.
        seqs = c(seqs, seqTxt)
        # Now move on to the next set of text lines.
        txt = c(lastSeqTxt, readLines(seqFile, nLinesAtOnce))
        }
    close(seqFile)
    if (length(txt)!= 0) stop("Unexpected text in fasta file: ", txt);
    return(seqs)
    }

#######################################################################################
# End of file.
#######################################################################################
Sourced.Include_TextFunctions = TRUE
cat("  Include_TextFunctions.R included\n")
