#######################################################################################
# This file contains R definitions and functions relating to the tomato introgression
# lines.  It shares tomato IL-related R code that may be useful to multiple lab members.
# Author: Ted Toal
# Lab: Brady
#######################################################################################

cat("Including Include_TomatoILsBILs.R\n")

#######################################################################################
# Tomato IL constants.
#######################################################################################

# Names of tomato parents and colors to use for them.
parents = c("M82", "PENN")
parentColors = c(M82="red", PENN="springgreen4")
parentColorsLight = c(M82="red", PENN="springgreen3")
parentColorNames = c(M82="red", PENN="green")

#######################################################################################
# Tomato IL constants.
#######################################################################################

# Tomato marker table file.
allTomatoMarkersFile = paste(TomatoILsBILsDir, "ITAG2.3_AllTomatoMarkerPositions.table", sep=PATHSEP)

# IL marker table file.
ILmarkersFile = paste(TomatoILsBILsDir, "ITAG2.3_IL_Markers.table", sep=PATHSEP)

# IL marker information table file.  This has more extensive info than above table.
ILmarkersInfoFile = paste(TomatoILsBILsDir, "ITAG2.3_IL_marker_info.table", sep=PATHSEP)

# IL marker region (e.g. 1A) table file.
ILmarkerRegionsFile = paste(TomatoILsBILsDir, "ITAG2.3_IL_region_info.table", sep=PATHSEP)

# IL region (e.g. 1A) table file.  This has more extensive info than above table.
ILregionsFile = paste(TomatoILsBILsDir, "ITAG2.3_IL_Regions.table", sep=PATHSEP)

# EXPEN 2000 marker table file.
markerPosFile = paste(TomatoILsBILsDir, "TomatoExpen2000MarkersWithPositions.table", sep=PATHSEP)

# Aasish's IL introgressions.
aashishIntrogressionFile = paste(TomatoILsBILsDir, "IL_introgressions_Aashish.tsv", sep=PATHSEP)

# Sinha lab IL files.
solyILboundaries = paste(TomatoILsBILsDir, "ITAG2.3_Sinha_Chitwood_Jul2012_ILbounds_Supplemental Table 2_edited.tsv", sep=PATHSEP) # From Aashish's
solyILbinMatrix = paste(TomatoILsBILsDir, "ITAG2.3_Sinha_Chitwood_Jul2012_binMatrix_Supplemental Dataset 2_edited.tsv", sep=PATHSEP)
solyILbinDefs = paste(TomatoILsBILsDir, "ITAG2.3_Sinha_Chitwood_Jul2012_binDefs_Supplemental Figures 5_6.tsv", sep=PATHSEP)
solyILbinBounds = paste(TomatoILsBILsDir, "ITAG2.3_Sinha_Chitwood_Jul2012_binBoundaries_derived.tsv", sep=PATHSEP)
solyILbinGenes = paste(TomatoILsBILsDir, "ITAG2.3_Sinha_Chitwood_Jul2012_binGenes_Supplemental Dataset 4.tsv", sep=PATHSEP)

# Sinha lab BIL files.
solyBILboundaries = paste(TomatoILsBILsDir, "ITAG2.3_Sinha_BIL_boundaries_24Jul2014_Incomplete.tsv", sep=PATHSEP)

# Brady lab RNAseq IL introgressions called from 2013 root RNAseq data.
RNAseq2013IntrogressionFile = paste(TomatoILsBILsDir, "ITAG2.3_IL_introgressions_RNAseq2013.tsv", sep=PATHSEP)

# Mixed Sinha/Brady IL introgressions, made by combining the solyILboundaries and
# RNAseq2013IntrogressionFile files, which cover slightly different sets of ILs.
# The first of these files uses Sinha data mostly, the second uses Brady lab data mostly.
solyILboundariesSinhaAndBrady = paste(TomatoILsBILsDir, "ITAG2.3_SinhaILs_20Aug2012_And_BradyILs_RNAseq2013.tsv", sep=PATHSEP)
solyILboundariesBradyAndSinha = paste(TomatoILsBILsDir, "ITAG2.3_BradyILs_RNAseq2013_And_SinhaILs_20Aug2012.tsv", sep=PATHSEP)

# The following files are remakes of Sinha-bin files, using solyILboundariesSinhaAndBrady
# for IL boundaries rather than solyILboundaries.  Prefix "bl" stands for "Brady Lab",
# and "SB" in the filename stands for "SinhaAndBradyILboundaries".
blBinMatrix = paste(TomatoILsBILsDir, "ITAG2.3_SB_binMatrix.tsv", sep=PATHSEP) # corresponds to solyILbinMatrix
blBinDefs = paste(TomatoILsBILsDir, "ITAG2.3_SB_binDefs.tsv", sep=PATHSEP) # corresponds to solyILbinDefs
blBinBounds = paste(TomatoILsBILsDir, "ITAG2.3_SB_binBounds.tsv", sep=PATHSEP) # corresponds to solyILbinBounds
blBinGenes = paste(TomatoILsBILsDir, "ITAG2.3_SB_binGenes.tsv", sep=PATHSEP) # corresponds to solyILbinGenes


#######################################################################################
# Tomato IL functions.
#######################################################################################

#######################################################################################
# Read Aashish introgression file and return a data frame of position of introgression
# for each IL.  Example:
#   intros = getIntrogressions()
# NOTE: This returns the "old" introgression positions, before it was known that some
# ILs have more than one introgression.
#######################################################################################
getIntrogressions = function()
    {
    aIntros = read.table(aashishIntrogressionFile, row.names=NULL, header=TRUE, stringsAsFactors=FALSE)
    rng = 1:(nrow(aIntros)-1)
    sameChrAsNext = aIntros$genotype[rng] == aIntros$genotype[rng+1] & aIntros$chr[rng] == aIntros$chr[rng+1]
    whichPENNstart = which(aIntros$parent == "PENN" & c(sameChrAsNext, FALSE))
    intros = aIntros[whichPENNstart,]
    intros = intros[,-4] # Remove "parent" column, don't need any more
    colnames(intros)[3] = "start" # Rename "pos" to "start"
    intros$end = aIntros[whichPENNstart+1,"pos"] - 1
    rownames(intros) = intros$genotype
    intros = intros[,-1] # Remove "genotype" as a column, it is row names
    return(intros)
    }

#######################################################################################
# Read Sinha introgression file and return a data frame of position of introgression
# for each IL.  Example:
#   intros = getNewIntrogressions()
#######################################################################################
getNewIntrogressions = function()
    {
    intros = read.table(solyILboundaries, sep="\t", header=TRUE, stringsAsFactors=FALSE)
    rownames(intros) = intros$IL_segment
    return(intros)
    }

#######################################################################################
# This function is like the function "order", and calls it with the ... arguments,
# returning its return value.  It differs in that it does special sorting on digits,
# the motivation being to sort a list of ILs such that e.g. IL10-1 sorts AFTER IL9-1-5.
# It adds leading zeroes to each string of 1 or more digits surrounded by non-digits
# so that all digit strings are the same length (the length of the longest digit string
# in any of the arguments), and sorting occurs on the modified values.  Thus, this can
# be used to sort many types of data, not just a list of ILs.
# E.g. to sort the IL-rows of a data frame df by "IL" then columns "chr" then "start":
#   ord = sortILs(rownames(df), df$chr, df$start)
#   df = df[ord,]
# As with order(), there can be more than one argument.  By default, only the first
# argument is modified to add leading zeroes to number strings.  The argument "lead0"
# can be specified to give the names of ... arguments to be modified with leading
# zeroes.
# E.g. to sort the rows of a data frame df by columns "trait" then "IL" then "pos":
#   ord = sortILs(df$trait, ILs=df$IL, df$pos, lead0="ILs")
#   df = df[ord,]
# Use the "first" and "last" arguments to list particular values which should sort to
# the start or end.  These arguments each are either a list or vector.  A vector is
# applied to the first ... argument and gives the values which are to be first or last
# in the sorted list.  A list value for "first" or "last" has as its elements vectors
# whose names are ... argument names, and the vectors contain the values within those
# argument names to be sorted first or last.
# E.g. to sort the rows of a data frame df by IL (with "M82" first and "PENN" last)
# and then by chr then start:
#   ord = sortILs(rownames(df), df$chr, df$start)
#   df = df[ord,]
# Arguments ... are passed to order() following the vector of (modified) IL names.
#######################################################################################
sortILs = function(..., lead0=NULL, first=c("M82", "PENN"), last=NULL)
    {
    return(order.special(..., lead0=lead0, first=first, last=last))
    }

#######################################################################################
# Test if a genomic position lies in the introgression of an IL.
#
# Return TRUE iff 1-based genetic position pos on chromosome chr is located within PENN
# genomic space based on introgression data intros (returned by getIntrogressions) for
# the genotype gt.  FALSE is always returned if gt == parents[1] = M82, and TRUE is
# always returned if gt == parents[2] = PENN.  Note: if pos and chr are vectors of
# length N, the return value is a logical vector of length N.
#######################################################################################
isInIntrogression = function(pos, chr, gt, intros)
    {
    if (gt == parents[1])
        return(rep(FALSE, length(pos)))
    if (gt == parents[2])
        return(rep(TRUE, length(pos)))
    if (!gt %in% rownames(intros)) { stop(paste("Bad genotype:", gt)) }
	return(chr == intros[gt,"chr"] & pos >= intros[gt,"start"] & pos <= intros[gt,"end"])
    }

#######################################################################################
# Get the distance of a genomic position from the introgression boundary of an IL.
#
# Return the distance, in Mb, of the 1-based genetic position pos on chromosome chr from
# the nearest introgression boundary on its chromosome.  If there is no introgression on
# its chromosome (as will always be true for parent genotypes), 0 is returned.  A
# positive number is returned if the position is located within the introgression, and a
# negative number if it is outside the introgression.  The value is rounded UP to the
# nearest larger (in absolute value) multiple of 1 Mb, so the distance is never 0 and 0 is
# reserved solely to mean there is no introgression on chromosome chr.  Note: if pos and
# chr are vectors of length N, the return value is a numeric vector of length N.
#######################################################################################
distFromIntroBoundary = function(pos, chr, gt, intros)
    {
    if (gt %in% parents)
        return(rep(0, length(pos)))
    if (!gt %in% rownames(intros)) { stop(paste("Bad genotype:", gt)) }
    Nstart = pos - intros[gt,"start"]
    Nend = pos - intros[gt,"end"]
    resultSign = myIfElse(sign(Nstart) == sign(Nend), -1, +1)
    Nstart = abs(Nstart)
    Nend = abs(Nend)
    N = pmin(Nstart, Nend)/1000000
    N[N == 0] = 1
    N = ceiling(N) * resultSign
    N[chr != intros[gt,"chr"]] = 0
	return(N)
    }

#######################################################################################
# Get lists of genes in specified (new Sinha) bins.
#
# Return a data frame with two columns: bin and gene.
#######################################################################################
getGenesInBins = function(bins)
    {
    geneBins = read.table(solyILbinGenes, header=TRUE, row.names=1, comment.char="", quote="\"", sep="\t", stringsAsFactors=FALSE)
    geneBins$gene = rownames(geneBins)
    geneBins = geneBins[, c("bin", "gene")]
    rownames(geneBins) = NULL
    geneBins = geneBins[geneBins$bin %in% bins,]
    return(geneBins)
    }

#######################################################################################
# End of file.
#######################################################################################
Sourced.Include_TomatoILsBILs = TRUE
cat("  Include_TomatoILsBILs.R included\n")
