#######################################################################################
# This file contains R definitions and functions relating to phenotype trait analyses.
# Author: Ted Toal
# Lab: Brady
#
# IMPORTANT:
#
# There have been three different experiments collecting phenotype data: (1) the one
# associated with the Sep 2011 RNA-seq experiment, where cell counts and length/angle
# data was measured; (2) the one associated with the Jun 2013 RNA-seq experiment,
# where root lengths and angles were measured; (3) the Spalding lab IL rotation
# experiment, where root tip angle was measured over a time course of about 34 time
# points.
#
# There have been multiple reanalyses of the phenotype data from some experiments.
# For example, the June 2013 experiment was reanalyzed with a different method of
# measuring angles involving dividing the root into thirds and measuring average
# absolute angle in each third.  And, the Spalding experiment was reanalyzed with
# separation of AWAY/THRU roots.  Data for each analysis is made available in this
# file.  The different analyses are identified by a keyword which is part of the
# name of the subdirectory where the analysis result data is stored.  Use functions
# getExperAnalysisInfo(), getCurExperAnalyses() and setCurExperAnalyses() to query
# and change the selected analyses and to see all the analysis keywords.  The first
# analysis of each experiment is always named "Orig", and the functions in this file
# default to using that analysis unless setCurExperAnalyses() is called to change
# it.  Therefore, the functions do not have an "analysis" argument to specify which
# analysis is to be used.
#
# The Spalding experiments present issues with subsets of traits.  In the "Orig"
# experiment the tip angle at each time point was treated as a trait, and the time
# derivative of the tip angles was treated as another set of traits, one at each
# time point.  In the AWAY/THRU analysis, each of these two sets of traits was
# subdivided further into ALL, AWAY, and THRU subsets of traits.  To handle this
# breakdown of measurements into different subsets of traits, a "subanalysis" element
# was added to this file.  Thus, the Spalding experiment has subanalyses named, e.g.
# tipAngle.AWAY and tipAngDeriv.THRU.  The other experiments currently do not have
# any subanalyses, so they are treated as having a single subanalysis named "All".
# (However, it would make sense to divide the data from the Sep 2011 experiment into
# two subanalyses, one for cell traits and one for length/angle traits, which would
# have the side-benefit of making it unnecessary to combine the results files from these
# two sets of measurements into single files, which currently is done by R code named
# CombineSummaryTables.R in the CellCountingGravity folder).  Use function getSubanalyses()
# to get a vector of subanalysis names.  The argument "subanalysisInExper" appears here
# in many functions.  It gives both the experiments and, within each experiment, the
# one or more subanalyses to be considered by the function.  The addition of
# subanalyses was a major change to this file that affects much other code that uses
# it.
#
# Argument subanalysisInExper: this argument to many of the functions in this file
# must be either a vector or list, or it may be NULL.  It describes which experiments
# and subanalyses the function is to be applied to.  If it is NULL, the function uses
# ALL experiments and ALL subanalyses (and typically will use the CURRENT analysis,
# although some functions may have an "analysis" argument).  If not NULL, each element
# NAME must be the name of an EXPERIMENT to which the function is to be applied.  The
# VALUE assigned to that element name must be one or more SUBANALYSIS names.  If just
# one, the vector form of this argument can be used, but if more than one, the list
# form must be used, since each list element must be a vector of subanalysis names.
# If a list is used, an experiment name may be assigned the value NULL, which means
# the function should use ALL subanalyses for that experiment.  The vector form of
# this argument cannot be used in that case because the experiment name is lost (try
# c(A=NULL) to see what happens!).
#
#######################################################################################

cat("Including Include_phenoAnalysis.R\n")

# Ensure required files have been sourced.
requireSourced("Include_phenoRNAseqRootSep2011", "Include_phenoAnalysis")

#######################################################################################
# Phenotype analysis directory paths.
#######################################################################################

# Directory "GrowthExperiments".
growthExperimentsDir = paste(BradyLabDir, "GrowthExperiments", sep=PATHSEP)

# Directory "GrowthExperiments/Tomato".
tomatoGrowthExperimentsDir = paste(growthExperimentsDir, "Tomato", sep=PATHSEP)

# Directory "AngleGeneAnalysis".
angleGeneDir = paste(BradyLabDir, "AngleGeneAnalysis", sep=PATHSEP)

#######################################################################################
# June 2013 RNAseq experiment.
#######################################################################################

# Folder for phenotype analysis of roots collected for the Jun 2013 RNAseq experiment.
phenoRNAseqRootJun2013Dir = myIfElse(exists("MY_phenoRNAseqRootJun2013Dir"),
    phenoRNAseqRootJun2013Dir, paste(tomatoGrowthExperimentsDir, "RootRNAseqJun2013_Phenotypes", sep=PATHSEP))

# Define date prefix, data subfolder name, data subfolder path, and data file prefix
# for each analysis.

# "Orig" is the original Jun 2013 RNAseq experiment phenotype analysis.
phenoRNAseqRootJun2013datePfx.Orig = "05Feb2015"
phenoRNAseqRootJun2013dataSubdir.Orig = paste("Analysis_", phenoRNAseqRootJun2013datePfx.Orig, "_Orig", sep="")
phenoRNAseqRootJun2013dataPath.Orig = paste(phenoRNAseqRootJun2013Dir, phenoRNAseqRootJun2013dataSubdir.Orig, sep=PATHSEP)
phenoRNAseqRootJun2013filePfx.Orig = paste(phenoRNAseqRootJun2013dataPath.Orig, phenoRNAseqRootJun2013datePfx.Orig, sep=PATHSEP)

# "RmvSamp" is a reanalysis that removes samples that were not harvested for RNAseq.
# This results in somewhat lower statistical power and a slight lowering of significant
# results, with a few genotypes in a few traits having reduced (or no) significant stars.
phenoRNAseqRootJun2013datePfx.RmvSamp = "18Mar2015"
phenoRNAseqRootJun2013dataSubdir.RmvSamp = paste("Analysis_", phenoRNAseqRootJun2013datePfx.RmvSamp, "_RmvSamp", sep="")
phenoRNAseqRootJun2013dataPath.RmvSamp = paste(phenoRNAseqRootJun2013Dir, phenoRNAseqRootJun2013dataSubdir.RmvSamp, sep=PATHSEP)
phenoRNAseqRootJun2013filePfx.RmvSamp = paste(phenoRNAseqRootJun2013dataPath.RmvSamp, phenoRNAseqRootJun2013datePfx.RmvSamp, sep=PATHSEP)

# "thirdsAngs" is a reanalysis that replaces the root angle measurements "tenthAng",
# "quarterAng" and "wholeAng" with measurements done like avgAbsAng but along each
# third of the root, and named "first3rdAng", "mid3rdAng", and "last3rdAng".
phenoRNAseqRootJun2013datePfx.thirdsAngs = "11May2015"
phenoRNAseqRootJun2013dataSubdir.thirdsAngs = paste("Analysis_", phenoRNAseqRootJun2013datePfx.thirdsAngs, "_thirdsAngs", sep="")
phenoRNAseqRootJun2013dataPath.thirdsAngs = paste(phenoRNAseqRootJun2013Dir, phenoRNAseqRootJun2013dataSubdir.thirdsAngs, sep=PATHSEP)
phenoRNAseqRootJun2013filePfx.thirdsAngs = paste(phenoRNAseqRootJun2013dataPath.thirdsAngs, phenoRNAseqRootJun2013datePfx.thirdsAngs, sep=PATHSEP)

# "thirdsAngB" is a reanalysis that simply adds trait groups first3rdAngNotTipAng and tipAngNotFirst3rdAng.
phenoRNAseqRootJun2013datePfx.thirdsAngB = "13Jul2015"
phenoRNAseqRootJun2013dataSubdir.thirdsAngB = paste("Analysis_", phenoRNAseqRootJun2013datePfx.thirdsAngB, "_thirdsAngB", sep="")
phenoRNAseqRootJun2013dataPath.thirdsAngB = paste(phenoRNAseqRootJun2013Dir, phenoRNAseqRootJun2013dataSubdir.thirdsAngB, sep=PATHSEP)
phenoRNAseqRootJun2013filePfx.thirdsAngB = paste(phenoRNAseqRootJun2013dataPath.thirdsAngB, phenoRNAseqRootJun2013datePfx.thirdsAngB, sep=PATHSEP)

#######################################################################################
# Spalding lab rotation experiment.
#
# IMPORTANT NOTE: There are two types of traits with phenotype data in this experiment,
# called "allILtipangles" and "allILderiv".  Both sets of data come from the same
# measurements; the second set is the time derivative of the first set.  Each set
# consists of several time points for each measured genotype.  The proper thing would
# probably have been to define some special functions here to access time course data
# as a series of time points, but that isn't what we did.  Instead, we treated each of
# the two sets of data as a separate "experiment", even though they were not.  Therefore,
# to access the two sets of data here, you must specify two different experiment names.
#######################################################################################

# Spalding lab folder.
spaldingDir = paste(angleGeneDir, "SpaldingLab", sep=PATHSEP)

# Define date prefix, data subfolder name, and separately for allILtipAngs and
# allILderiv, the data subfolder path and data file prefix for each analysis.

# "Orig" is the original Spalding lab rotation experiment.
spaldingRotationDatePfx.Orig = "05Feb2015"
spaldingRotationDataSubdir.Orig = paste(spaldingRotationDatePfx.Orig, "_Orig", sep="")
spaldingRotationDataPath.Orig = paste(spaldingDir, "Analysis_Rot", spaldingRotationDataSubdir.Orig, sep=PATHSEP)
spaldingRotationFilePfx.Orig = paste(spaldingRotationDataPath.Orig, spaldingRotationDatePfx.Orig, sep=PATHSEP)
spalding_tipAngDerivsFilePfx.Orig = paste(spaldingRotationFilePfx.Orig, "_allILderiv_min4", sep="")
spalding_tipAnglesFilePfx.Orig = paste(spaldingRotationFilePfx.Orig, "_allILtipangles_min4", sep="")

# "AwayThru" is a reanalysis that separates the root rotation directions AWAY and
# THRU from each other.  There is a separate data file prefix for ALL, AWAY, and THRU
# subanalyses.
spaldingRotationDatePfx.AwayThru = "21May2015"
spaldingRotationDataSubdir.AwayThru = paste(spaldingRotationDatePfx.AwayThru, "_AwayThru", sep="")
spaldingRotationDataPath.AwayThru = paste(spaldingDir, "Analysis_Rot", spaldingRotationDataSubdir.AwayThru, sep=PATHSEP)
spaldingRotationFilePfx.AwayThru = paste(spaldingRotationDataPath.AwayThru, spaldingRotationDatePfx.AwayThru, sep=PATHSEP)
spalding_tipAngDerivs_ALL_FilePfx.AwayThru = paste(spaldingRotationFilePfx.AwayThru, "_allILderiv_ALL_min4", sep="")
spalding_tipAngDerivs_AWAY_FilePfx.AwayThru = paste(spaldingRotationFilePfx.AwayThru, "_allILderiv_AWAY_min3", sep="")
spalding_tipAngDerivs_THRU_FilePfx.AwayThru = paste(spaldingRotationFilePfx.AwayThru, "_allILderiv_THRU_min3", sep="")
spalding_tipAngles_ALL_FilePfx.AwayThru = paste(spaldingRotationFilePfx.AwayThru, "_allILtipangles_ALL_min4", sep="")
spalding_tipAngles_AWAY_FilePfx.AwayThru = paste(spaldingRotationFilePfx.AwayThru, "_allILtipangles_AWAY_min3", sep="")
spalding_tipAngles_THRU_FilePfx.AwayThru = paste(spaldingRotationFilePfx.AwayThru, "_allILtipangles_THRU_min3", sep="")

#######################################################################################
# Include the phenoRNAseqRootSep2011 phenotype include file, unless Include_RootRsourceFile.R
# is being used.
#######################################################################################

if (!exists("Sourced.Include_RootRsourceFile"))
    source(Include_phenoRNAseqRootSep2011)

#######################################################################################
# Lists collecting together the file paths and names for all phenotype analyses.
#######################################################################################

# Each analysis of a set of traits will produce several files that we want to provide
# paths to here.  Each file is identified by a keyword.  Additional keywords associated
# with each analysis may provide static information about the analysis (e.g. "desc").
# The keywords and their descriptions are as follows:
#   desc: description of the analysis and subanalysis (not of the experiment).
#   shortEAS: a short (unique) name that may be used for the experiment/analysis/subanalysis
#               names all combined.
#   shortTraitPfx: a short (unique) name that may be used as a prefix prepended to any
#               trait name to obtain a name that is unique across all subanalysis/trait
#               combinations (but not unique across all experiment/analysis/subanalysis/
#               trait combinations).
#   traits: name of file containing units and descriptions of phenotype data, with two
#               columns, "unit" and "desc", and with row names equal to trait names.
#   groups: name of file containing grouping of related traits, with columns "group"
#               and "formula".
#   mean: name of file containing mean phenotype value, with three or four columns:
#               "trait", "genotype", "mean", and "isLSmean" (the "isLSmean" column is
#               optional).
#   stderr: name of file containing standard error of the mean of phenotype values of
#               each trait in each IL, with IL rownames and columns being trait names.
#   pvalue: name of file containing p-values for comparison of each trait in each IL
#               to M82, with IL rownames and columns being trait names.
#   plus:  name of file containing number of "plusses" (aka "stars") assigned to each
#               trait in each IL based on the p-value, with IL rownames and columns
#               being trait names. (1 = + = 0.05, 2 = ++ = 0.01, 3 = +++ = 0.001)
#   qvalue: name of file containing q-values = p-values adjusted for multiple testing,
#               like pvalue.
#   stars: name of file containing number of stars assigned to each trait in each IL
#               based on the q-value, with IL rownames and columns being trait names.
#                   (1 = * = 0.05, 2 = ** = 0.01, 3 = *** = 0.001)
#   biorepMeans: name of file containing the mean of each trait, for each biorep within
#               each genotype, with these columns: "genotype", "biorep", and one column
#               for each trait.  Not all experiments produce this data, use NULL below
#               for the filename if not.

# Define a vector listing the phenotype data keywords that are associated with a file,
# and a second vector listing only the keywords for data files that have values for
# each IL.
data.AllFileKeywords = c("traits", "groups", "mean", "stderr", "pvalue", "plus", "qvalue", "stars", "biorepMeans")
data.ILkeywords = c("mean", "stderr", "pvalue", "plus", "qvalue", "stars")

# Define a list for each analysis, containing the above members, with the list element
# name being the data keyword.  Some analyses may be missing some files, in which case
# the list entry is NULL.

###########################################
# RNAseqRootSep2011 phenotyping experiment.
###########################################

######################
# "Orig" analysis.
######################

# There is currently only one analysis, which we will call "Orig".  Its paths
# have already been defined as variables "traitsFile", etc.  We should change
# this and get rid of those, making this like the others below.
phenoData.RNAseqRootSep2011.Orig = list(
    desc="Original phenotype data analysis for this experiment",
    shortEAS="Root2011.Orig",
    shortTraitPfx="", # No prefix necessary for trait name uniqueness.
    traits=traitsFile,
    groups=traitGroupsFile,
    mean=traitMeansFile,
    stderr=traitStdErrFile,
    pvalue=traitPvalsFile,
    plus=traitPlusFile,
    qvalue=traitQvalsFile,
    stars=traitStarFile,
    biorepMeans=NULL)

###########################################
# RNAseqRootJun2013 phenotyping experiment.
###########################################

######################
# "Orig" analysis.
######################

phenoData.RNAseqRootJun2013.Orig = list(
    desc="Original phenotype data analysis for this experiment",
    shortEAS="Root2013.Orig",
    shortTraitPfx="", # No prefix necessary for trait name uniqueness.
    traits=paste(phenoRNAseqRootJun2013filePfx.Orig, "traits.tsv", sep="_"),
    groups=paste(phenoRNAseqRootJun2013filePfx.Orig, "traitGroups.tsv", sep="_"),
    mean=paste(phenoRNAseqRootJun2013filePfx.Orig, "mean.tsv", sep="_"),
    stderr=paste(phenoRNAseqRootJun2013filePfx.Orig, "sem.tsv", sep="_"),
    pvalue=paste(phenoRNAseqRootJun2013filePfx.Orig, "p-value.tsv", sep="_"),
    plus=paste(phenoRNAseqRootJun2013filePfx.Orig, "plus.tsv", sep="_"),
    qvalue=paste(phenoRNAseqRootJun2013filePfx.Orig, "q-value.tsv", sep="_"),
    stars=paste(phenoRNAseqRootJun2013filePfx.Orig, "stars.tsv", sep="_"),
    biorepMeans=paste(phenoRNAseqRootJun2013filePfx.Orig, "biorepMeans.tsv", sep="_"))

######################
# "RmvSamp" analysis.
######################

phenoData.RNAseqRootJun2013.RmvSamp = list(
    desc="Analysis removes samples not collected for RNA extraction",
    shortEAS="Root2013.RmvSamp",
    shortTraitPfx="", # No prefix necessary for trait name uniqueness.
    traits=paste(phenoRNAseqRootJun2013filePfx.RmvSamp, "traits.tsv", sep="_"),
    groups=paste(phenoRNAseqRootJun2013filePfx.RmvSamp, "traitGroups.tsv", sep="_"),
    mean=paste(phenoRNAseqRootJun2013filePfx.RmvSamp, "mean.tsv", sep="_"),
    stderr=paste(phenoRNAseqRootJun2013filePfx.RmvSamp, "sem.tsv", sep="_"),
    pvalue=paste(phenoRNAseqRootJun2013filePfx.RmvSamp, "p-value.tsv", sep="_"),
    plus=paste(phenoRNAseqRootJun2013filePfx.RmvSamp, "plus.tsv", sep="_"),
    qvalue=paste(phenoRNAseqRootJun2013filePfx.RmvSamp, "q-value.tsv", sep="_"),
    stars=paste(phenoRNAseqRootJun2013filePfx.RmvSamp, "stars.tsv", sep="_"),
    biorepMeans=paste(phenoRNAseqRootJun2013filePfx.RmvSamp, "biorepMeans.tsv", sep="_"))

######################
# "thirdsAngs" analysis.
######################

phenoData.RNAseqRootJun2013.thirdsAngs = list(
    desc="Analysis replaces tenthAng/quarterAng/wholeAng with first3rdAng/mid3rdAng/last3rdAng",
    shortEAS="Root2013.thirds",
    shortTraitPfx="", # No prefix necessary for trait name uniqueness.
    traits=paste(phenoRNAseqRootJun2013filePfx.thirdsAngs, "traits.tsv", sep="_"),
    groups=paste(phenoRNAseqRootJun2013filePfx.thirdsAngs, "traitGroups.tsv", sep="_"),
    mean=paste(phenoRNAseqRootJun2013filePfx.thirdsAngs, "mean.tsv", sep="_"),
    stderr=paste(phenoRNAseqRootJun2013filePfx.thirdsAngs, "sem.tsv", sep="_"),
    pvalue=paste(phenoRNAseqRootJun2013filePfx.thirdsAngs, "p-value.tsv", sep="_"),
    plus=paste(phenoRNAseqRootJun2013filePfx.thirdsAngs, "plus.tsv", sep="_"),
    qvalue=paste(phenoRNAseqRootJun2013filePfx.thirdsAngs, "q-value.tsv", sep="_"),
    stars=paste(phenoRNAseqRootJun2013filePfx.thirdsAngs, "stars.tsv", sep="_"),
    biorepMeans=paste(phenoRNAseqRootJun2013filePfx.thirdsAngs, "biorepMeans.tsv", sep="_"))

######################
# "thirdsAngB" analysis.
######################

phenoData.RNAseqRootJun2013.thirdsAngB = list(
    desc="Analysis adds trait groups tipAngNotFirst3rdAng and first3rdAngNotTipAng",
    shortEAS="Root2013.3rdsB",
    shortTraitPfx="", # No prefix necessary for trait name uniqueness.
    traits=paste(phenoRNAseqRootJun2013filePfx.thirdsAngB, "traits.tsv", sep="_"),
    groups=paste(phenoRNAseqRootJun2013filePfx.thirdsAngB, "traitGroups.tsv", sep="_"),
    mean=paste(phenoRNAseqRootJun2013filePfx.thirdsAngB, "mean.tsv", sep="_"),
    stderr=paste(phenoRNAseqRootJun2013filePfx.thirdsAngB, "sem.tsv", sep="_"),
    pvalue=paste(phenoRNAseqRootJun2013filePfx.thirdsAngB, "p-value.tsv", sep="_"),
    plus=paste(phenoRNAseqRootJun2013filePfx.thirdsAngB, "plus.tsv", sep="_"),
    qvalue=paste(phenoRNAseqRootJun2013filePfx.thirdsAngB, "q-value.tsv", sep="_"),
    stars=paste(phenoRNAseqRootJun2013filePfx.thirdsAngB, "stars.tsv", sep="_"),
    biorepMeans=paste(phenoRNAseqRootJun2013filePfx.thirdsAngB, "biorepMeans.tsv", sep="_"))

###########################################
# ESrot Spalding Lab rotation experiment.
###########################################

######################
# "Orig" analysis.
######################

# "tipAngles" subanalysis.
phenoData.ESrot.Orig.tipAngles = list(
    desc="Original tip angle subanalysis for this experiment",
    shortEAS="ESorig.tipAngs",
    shortTraitPfx="tipAng",
    traits=paste(spalding_tipAnglesFilePfx.Orig, "traits.tsv", sep="_"),
    groups=paste(spalding_tipAnglesFilePfx.Orig, "traitGroups.tsv", sep="_"),
    mean=paste(spalding_tipAnglesFilePfx.Orig, "mean.tsv", sep="_"),
    stderr=paste(spalding_tipAnglesFilePfx.Orig, "sem.tsv", sep="_"),
    pvalue=paste(spalding_tipAnglesFilePfx.Orig, "p-value.tsv", sep="_"),
    plus=paste(spalding_tipAnglesFilePfx.Orig, "plus.tsv", sep="_"),
    qvalue=paste(spalding_tipAnglesFilePfx.Orig, "q-value.tsv", sep="_"),
    stars=paste(spalding_tipAnglesFilePfx.Orig, "stars.tsv", sep="_"),
    biorepMeans=NULL)

# "tipAngDerivs" subanalysis.
phenoData.ESrot.Orig.tipAngDerivs = list(
    desc="Original tip angle derivative subanalysis for this experiment",
    shortEAS="ESorig.angDerivs",
    shortTraitPfx="deriv",
    traits=paste(spalding_tipAngDerivsFilePfx.Orig, "traits.tsv", sep="_"),
    groups=paste(spalding_tipAngDerivsFilePfx.Orig, "traitGroups.tsv", sep="_"),
    mean=paste(spalding_tipAngDerivsFilePfx.Orig, "mean.tsv", sep="_"),
    stderr=paste(spalding_tipAngDerivsFilePfx.Orig, "sem.tsv", sep="_"),
    pvalue=paste(spalding_tipAngDerivsFilePfx.Orig, "p-value.tsv", sep="_"),
    plus=paste(spalding_tipAngDerivsFilePfx.Orig, "plus.tsv", sep="_"),
    qvalue=paste(spalding_tipAngDerivsFilePfx.Orig, "q-value.tsv", sep="_"),
    stars=paste(spalding_tipAngDerivsFilePfx.Orig, "stars.tsv", sep="_"),
    biorepMeans=NULL)

######################
# "AwayThru" analysis.
######################

# "tipAngles.ALL" subanalysis.
phenoData.ESrot.AwayThru.tipAngles.ALL = list(
    desc="Combined AWAY/THRU tip angle subanalysis in AWAY/THRU analysis",
    shortEAS="ESalt.tipAngsALL",
    shortTraitPfx="ALLtipAng",
    traits=paste(spalding_tipAngles_ALL_FilePfx.AwayThru, "traits.tsv", sep="_"),
    groups=paste(spalding_tipAngles_ALL_FilePfx.AwayThru, "traitGroups.tsv", sep="_"),
    mean=paste(spalding_tipAngles_ALL_FilePfx.AwayThru, "mean.tsv", sep="_"),
    stderr=paste(spalding_tipAngles_ALL_FilePfx.AwayThru, "sem.tsv", sep="_"),
    pvalue=paste(spalding_tipAngles_ALL_FilePfx.AwayThru, "p-value.tsv", sep="_"),
    plus=paste(spalding_tipAngles_ALL_FilePfx.AwayThru, "plus.tsv", sep="_"),
    qvalue=paste(spalding_tipAngles_ALL_FilePfx.AwayThru, "q-value.tsv", sep="_"),
    stars=paste(spalding_tipAngles_ALL_FilePfx.AwayThru, "stars.tsv", sep="_"),
    biorepMeans=NULL)

# "tipAngles.AWAY" subanalysis.
phenoData.ESrot.AwayThru.tipAngles.AWAY = list(
    desc="AWAY tip angle subanalysis in AWAY/THRU analysis",
    shortEAS="ESalt.tipAngsAWAY",
    shortTraitPfx="AWAYtipAng",
    traits=paste(spalding_tipAngles_AWAY_FilePfx.AwayThru, "traits.tsv", sep="_"),
    groups=paste(spalding_tipAngles_AWAY_FilePfx.AwayThru, "traitGroups.tsv", sep="_"),
    mean=paste(spalding_tipAngles_AWAY_FilePfx.AwayThru, "mean.tsv", sep="_"),
    stderr=paste(spalding_tipAngles_AWAY_FilePfx.AwayThru, "sem.tsv", sep="_"),
    pvalue=paste(spalding_tipAngles_AWAY_FilePfx.AwayThru, "p-value.tsv", sep="_"),
    plus=paste(spalding_tipAngles_AWAY_FilePfx.AwayThru, "plus.tsv", sep="_"),
    qvalue=paste(spalding_tipAngles_AWAY_FilePfx.AwayThru, "q-value.tsv", sep="_"),
    stars=paste(spalding_tipAngles_AWAY_FilePfx.AwayThru, "stars.tsv", sep="_"),
    biorepMeans=NULL)

# "tipAngles.THRU" subanalysis.
phenoData.ESrot.AwayThru.tipAngles.THRU = list(
    desc="THRU tip angle subanalysis in AWAY/THRU analysis",
    shortEAS="ESalt.tipAngsTHRU",
    shortTraitPfx="THRUtipAng",
    traits=paste(spalding_tipAngles_THRU_FilePfx.AwayThru, "traits.tsv", sep="_"),
    groups=paste(spalding_tipAngles_THRU_FilePfx.AwayThru, "traitGroups.tsv", sep="_"),
    mean=paste(spalding_tipAngles_THRU_FilePfx.AwayThru, "mean.tsv", sep="_"),
    stderr=paste(spalding_tipAngles_THRU_FilePfx.AwayThru, "sem.tsv", sep="_"),
    pvalue=paste(spalding_tipAngles_THRU_FilePfx.AwayThru, "p-value.tsv", sep="_"),
    plus=paste(spalding_tipAngles_THRU_FilePfx.AwayThru, "plus.tsv", sep="_"),
    qvalue=paste(spalding_tipAngles_THRU_FilePfx.AwayThru, "q-value.tsv", sep="_"),
    stars=paste(spalding_tipAngles_THRU_FilePfx.AwayThru, "stars.tsv", sep="_"),
    biorepMeans=NULL)

# "tipAngDerivs.ALL" subanalysis.
phenoData.ESrot.AwayThru.tipAngDerivs.ALL = list(
    desc="Combined AWAY/THRU tip angle derivative subanalysis in AWAY/THRU analysis",
    shortEAS="ESalt.angDerivsALL",
    shortTraitPfx="ALLderiv",
    traits=paste(spalding_tipAngDerivs_ALL_FilePfx.AwayThru, "traits.tsv", sep="_"),
    groups=paste(spalding_tipAngDerivs_ALL_FilePfx.AwayThru, "traitGroups.tsv", sep="_"),
    mean=paste(spalding_tipAngDerivs_ALL_FilePfx.AwayThru, "mean.tsv", sep="_"),
    stderr=paste(spalding_tipAngDerivs_ALL_FilePfx.AwayThru, "sem.tsv", sep="_"),
    pvalue=paste(spalding_tipAngDerivs_ALL_FilePfx.AwayThru, "p-value.tsv", sep="_"),
    plus=paste(spalding_tipAngDerivs_ALL_FilePfx.AwayThru, "plus.tsv", sep="_"),
    qvalue=paste(spalding_tipAngDerivs_ALL_FilePfx.AwayThru, "q-value.tsv", sep="_"),
    stars=paste(spalding_tipAngDerivs_ALL_FilePfx.AwayThru, "stars.tsv", sep="_"),
    biorepMeans=NULL)

# "tipAngDerivs.AWAY" subanalysis.
phenoData.ESrot.AwayThru.tipAngDerivs.AWAY = list(
    desc="Combined AWAY/THRU tip angle derivative subanalysis in AWAY/THRU analysis",
    shortEAS="ESalt.angDerivsAWAY",
    shortTraitPfx="AWAYderiv",
    traits=paste(spalding_tipAngDerivs_AWAY_FilePfx.AwayThru, "traits.tsv", sep="_"),
    groups=paste(spalding_tipAngDerivs_AWAY_FilePfx.AwayThru, "traitGroups.tsv", sep="_"),
    mean=paste(spalding_tipAngDerivs_AWAY_FilePfx.AwayThru, "mean.tsv", sep="_"),
    stderr=paste(spalding_tipAngDerivs_AWAY_FilePfx.AwayThru, "sem.tsv", sep="_"),
    pvalue=paste(spalding_tipAngDerivs_AWAY_FilePfx.AwayThru, "p-value.tsv", sep="_"),
    plus=paste(spalding_tipAngDerivs_AWAY_FilePfx.AwayThru, "plus.tsv", sep="_"),
    qvalue=paste(spalding_tipAngDerivs_AWAY_FilePfx.AwayThru, "q-value.tsv", sep="_"),
    stars=paste(spalding_tipAngDerivs_AWAY_FilePfx.AwayThru, "stars.tsv", sep="_"),
    biorepMeans=NULL)

# "tipAngDerivs.THRU" subanalysis.
phenoData.ESrot.AwayThru.tipAngDerivs.THRU = list(
    desc="Combined AWAY/THRU tip angle derivative subanalysis in AWAY/THRU analysis",
    shortEAS="ESalt.angDerivsTHRU",
    shortTraitPfx="THRUderiv",
    traits=paste(spalding_tipAngDerivs_THRU_FilePfx.AwayThru, "traits.tsv", sep="_"),
    groups=paste(spalding_tipAngDerivs_THRU_FilePfx.AwayThru, "traitGroups.tsv", sep="_"),
    mean=paste(spalding_tipAngDerivs_THRU_FilePfx.AwayThru, "mean.tsv", sep="_"),
    stderr=paste(spalding_tipAngDerivs_THRU_FilePfx.AwayThru, "sem.tsv", sep="_"),
    pvalue=paste(spalding_tipAngDerivs_THRU_FilePfx.AwayThru, "p-value.tsv", sep="_"),
    plus=paste(spalding_tipAngDerivs_THRU_FilePfx.AwayThru, "plus.tsv", sep="_"),
    qvalue=paste(spalding_tipAngDerivs_THRU_FilePfx.AwayThru, "q-value.tsv", sep="_"),
    stars=paste(spalding_tipAngDerivs_THRU_FilePfx.AwayThru, "stars.tsv", sep="_"),
    biorepMeans=NULL)

#######################################################################################
# Lists, one per experiment, that collect together all of the above lists for all
# phenotype analyses, and defining the subanalyses of each analysis.  For most
# experiments the sub-list elements here are trivial, but the Spalding experiment
# is different because it is a time course and each time course point was treated as
# a trait, necessitating calling as "subanalyses" what would otherwise have been
# called "traits".
# Examples of subanalyses: tipAngles, tipAngDerivs, tipAngles.AWAY.
# The structure of each list here is:
#   <experiment name>=list(
#       <analysis name>=list(
#           <subanalysis name>=<above list of files in this subanalysis>,
#           ...
#       )
#   )
#
# IMPORTANT NOTE: For experiments that have only one subanalysis, the subanalysis name
# is "All".
#######################################################################################

# Master list of all analyses for RNAseqRootSep2011 experiment.
phenoData.RNAseqRootSep2011 = list(Orig=list(all=phenoData.RNAseqRootSep2011.Orig))

# Master list of all analyses for RNAseqRootSep2011 experiment.
phenoData.RNAseqRootJun2013 = list(
    Orig=list(all=phenoData.RNAseqRootJun2013.Orig),
    RmvSamp=list(all=phenoData.RNAseqRootJun2013.RmvSamp),
    thirdsAngs=list(all=phenoData.RNAseqRootJun2013.thirdsAngs),
    thirdsAngB=list(all=phenoData.RNAseqRootJun2013.thirdsAngB)
    )

# Master list of all analyses for Spalding Lab rotation experiment.
phenoData.ESrot = list(
    Orig=list(
        tipAngles=phenoData.ESrot.Orig.tipAngles,
        tipAngDerivs=phenoData.ESrot.Orig.tipAngDerivs),
    AwayThru=list(
        tipAngles.ALL=phenoData.ESrot.AwayThru.tipAngles.ALL,
        tipAngles.AWAY=phenoData.ESrot.AwayThru.tipAngles.AWAY,
        tipAngles.THRU=phenoData.ESrot.AwayThru.tipAngles.THRU,
        tipAngDerivs.ALL=phenoData.ESrot.AwayThru.tipAngDerivs.ALL,
        tipAngDerivs.AWAY=phenoData.ESrot.AwayThru.tipAngDerivs.AWAY,
        tipAngDerivs.THRU=phenoData.ESrot.AwayThru.tipAngDerivs.THRU)
    )

#######################################################################################
# Finally, a list containing all of the above experiment lists.  The names of the list
# elements are experiment names.
#######################################################################################

all.phenoData = list(
    RNAseqRootSep2011=phenoData.RNAseqRootSep2011,
    RNAseqRootJun2013=phenoData.RNAseqRootJun2013,
    ESrot=phenoData.ESrot)

#######################################################################################
# Experiment descriptions and citations.  The names are experiment names.
#######################################################################################
experimentInfo = list(
    RNAseqRootSep2011=list(
        desc="Measure tomato/pennellii IL and parent primary root cell, length, and angle traits at 5 to 7 days after germination",
        citation="Ron, Mily et al. Identification of novel loci regulating interspecific variation in root morphology and cellular development in tomato. Plant physiology (2013)"),
    RNAseqRootJun2013=list(
        desc="Measure tomato/pennellii IL and parent primary root length and angle traits and perform RNA-seq on root tip tissue at 3 to 5 days after germination",
        citation="(in progress)"),
    ESrot=list(
        desc="Measure tomato/pennellii IL and parent primary root tip angle and derivative after 90 degree rotation",
        citation="(in progress)"))

#######################################################################################
# Currently selected analysis for each experiment.  Initialize all experiments to use
# analysis "Orig".
#######################################################################################
curAnalysis = c(
    RNAseqRootSep2011="Orig",
    RNAseqRootJun2013="Orig",
    ESrot="Orig")

#######################################################################################
# General functions for phenotype analysis.
#######################################################################################

#######################################################################################
# Given arguments that are vectors of names of experiments, subanalyses, groups,
# or traits, paste them together using "-" as a separator.  This character is not
# used within any experiment, subanalysis, group, or trait name ("." or "_" is
# used in those names if they need word separation), and it is a valid tab name
# in Excel.  By pasting them together with "-", they can be taken apart later if
# need be, using function split_ExperNames().  This is useful when these combined
# names are used as tab names in an Excel file.
#
# Arguments:
#   ...: Two or more arguments that are vectors, all the same size, of names of
#           experiments, subanalyses, groups, or traits, to be pasted together.
#
# Returns: Vector of same length as argument vectors, containing pasted-together names.
#######################################################################################
join_ExperNames = function(...)
    {
    L = list(...)
    L[["sep"]] = "-"
    V = do.call(paste, L)
    return(V)
    }

#######################################################################################
# Given a string that consists of experiments, subanalyses, groups, or traits that
# were pasted together using "-" as a separator (e.g. by join_ExperNames()), split
# them apart.
#
# Arguments:
#   S: vector of strings, each joined-togther names of experiments, subanalyses,
#       groups, or traits.
#   simplify: if TRUE and the S argument is a vector of length 1, a vector containing
#       the split elements is returned.  Otherwise, a list of length equal to the
#       length of S is returned, each element a vector containing the split elements.
#
# Returns: List or vector of split names, see "simplify" argument.
#######################################################################################
split_ExperNames = function(S, simplify=TRUE)
    {
    L = strsplit(S, "-", fixed=TRUE)
    if (simplify && length(L) == 1)
        return(unlist(L))
    return(L)
    }

#######################################################################################
# Given a vector of subanalysis names, with the element names being experiment names
# in the manner of the "subanalysisInExper" arguments used by many functions herein,
# combine the experiment and subanalysis names together into a single name separated
# by "_".
#
# This is useful when these combined names are used as tab names in an Excel file.
#
# Arguments:
#   subanalysisInExper: see comments at top of file.
#
# Returns: a vector of the combined names.  Use convertTo_subanalysisInExper() to
# convert the result back again.
#######################################################################################
convertTo_exper.sub = function(subanalysisInExper)
    {
    return(join_ExperNames(names(subanalysisInExper), subanalysisInExper))
    }

#######################################################################################
# Given a combined experiment/subanalysis name containing one "-" character separating
# the experiment name from the subanalysis name, break the two names apart and create
# a vector whose value is the subanalysis name and the name of the value is the
# experiment name.
#
# Arguments:
#   exper.sub: vector of joined-togther experiment and subanalysis names to be separated.
#
# Returns: A vector suitable for the subanalysisInExper argument to many of the
# functions herein.  The argument may be a vector of exper.sub names, in which case
# the returned vector has one entry for each.
#######################################################################################
convertTo_subanalysisInExper = function(exper.sub)
    {
    L = strsplit(exper.sub, "-", fixed=TRUE)
    expers = sapply(L, function(V) V[1])
    subanalyses = sapply(L, function(V) V[2])
    names(subanalyses) = expers
    return(subanalyses)
    }

#######################################################################################
# Check for existence and integrity of all data files of all analyses of all experiments
# by reading them, and display a message for missing ones.
#######################################################################################
checkPhenoDataFiles = function()
    {
    Nfiles = 0
    for (exper in names(all.phenoData))
        {
        #cat(" experiment:", exper, "\n")
        analyses = names(all.phenoData[[exper]])
        for (analysis in analyses)
            {
            #cat("  analysis:", analysis, "\n")
            subanalyses = names(all.phenoData[[exper]][[analysis]])
            for (subanalysis in subanalyses)
                {
                #cat("   subanalysis:", subanalysis, "\n")
                for (keyword in data.AllFileKeywords)
                    {
                    #cat("    keyword:", keyword, "\n")
                    dataFile = all.phenoData[[exper]][[analysis]][[subanalysis]][[keyword]]
                    #cat("dataFile = ", dataFile, "\n")
                    if (is.null(dataFile))
                        next
                    x = tryCatch(
                        df <- read.table(dataFile, sep="\t", header=TRUE, row.names=NULL, stringsAsFactors=FALSE),
                        warning=function(w) w, error=function(e) e)
                    if (is(x, "warning") || is(x, "error"))
                        stop("Experiment:", exper, " analysis:", analysis,
                            " subanalysis:", subanalysis, " keyword:", keyword, " Error: ", x$message)
                    Nfiles = Nfiles + 1
                    }
                }
            }
        }
    cat(Nfiles, " phenotype data files appear to be intact.\n")
    }

#######################################################################################
# Return a vector of experiment names that are available.
#######################################################################################
getAllExperimentNames = function()
    {
    return(names(all.phenoData))
    }

#######################################################################################
# Return a data frame of experiment information for the specified phenotype experiments.
#
# Arguments:
#   experiments: a vector of experiment names for which experiment information is desired,
#       or NULL to return information for ALL experiments.
#
# Returns: data frame with experiments as row names, and with columns "desc" and "citation".
#######################################################################################
getExperimentInfo = function(experiments=NULL)
    {
    allExps = getAllExperimentNames()
    if (is.null(experiments))
        experiments = allExps
    experiments = unique(experiments)
    badExps = !experiments %in% allExps
    if (any(badExps))
        stop("getExperimentInfo: unknown experiment name(s): ",
            paste(experiments[badExps], collapse=", "))

    df = data.frame(desc=rep(NA, length(experiments)), citation=NA, row.names=experiments, stringsAsFactors=FALSE)
    for (exper in experiments)
        {
        df[exper, "desc"] = experimentInfo[[exper]]$desc
        df[exper, "citation"] = experimentInfo[[exper]]$citation
        }
    return(df)
    }

#######################################################################################
# Return a data frame of experiment ANALYSIS and SUBANALYSIS information for the
# specified experiments, analyses, and subanalyses.
#
# Arguments:
#   subanalysisInExper: see comments at top of file.
#   analyses: a vector of analysis names (any experiment) for which analysis and
#       subanalysis information is desired, or NULL to return information for ALL
#       analyses.
#
# Returns: data frame with columns "exper", "analysis", "subanalysis", "desc",
# "shortEAS", "shortTraitPfx", and "cur" containing an experiment name, analysis
# name, subanalysis name, analysis and subanalysis description, short name for
# the experiment/analysis/subanalysis, short trait prefix name for the subanalysis,
# and current analysis TRUE/FALSE flag.
#######################################################################################
getExperAnalysisInfo = function(subanalysisInExper=NULL, analyses=NULL)
    {
    dfExper = getExperimentInfo(names(subanalysisInExper)) # names(NULL) is NULL
    experiments = rownames(dfExper)
    df = NULL
    for (exper in experiments)
        {
        exper.analyses = names(all.phenoData[[exper]])
        if (!is.null(analyses))
            exper.analyses = intersect(exper.analyses, analyses)
        for (analysis in exper.analyses)
            {
            analysis.subanalyses = names(all.phenoData[[exper]][[analysis]])
            if (!is.null(subanalysisInExper[[exper]])) # NULL[["x"]] is NULL
                analysis.subanalyses = intersect(analysis.subanalyses, subanalysisInExper[[exper]])
            for (subanalysis in analysis.subanalyses)
                {
                dfPhenoData = all.phenoData[[exper]][[analysis]][[subanalysis]]
                df = rbind(df, data.frame(exper=exper, analysis=analysis, subanalysis=subanalysis,
                    desc=dfPhenoData$desc, shortEAS=dfPhenoData$shortEAS,
                    shortTraitPfx=dfPhenoData$shortTraitPfx, cur=(analysis == curAnalysis[exper]),
                    row.names=NULL, stringsAsFactors=FALSE))
                }
            }
        }
    return(df)
    }

#######################################################################################
# Return the CURRENT analysis to be used for each of the specified experiments.
#
# Returns: a vector of analysis names, with vector element names being experiment
# names and vector element values being the names of the current analyses used for
# those experiments.
#######################################################################################
getCurExperAnalyses = function()
    {
    return(curAnalysis)
    }

#######################################################################################
# Set the CURRENT analysis to be used for each of the specified experiments.
#
# Arguments:
#   analyses: a vector of analysis names with vector element names being experiment
#       names and vector element values being the names of the analyses to be set as the
#       current ones for those experiments.  May not be NULL.
#
# Returns: nothing.
#######################################################################################
setCurExperAnalyses = function(analyses)
    {
    if (is.null(analyses) || !all(names(analyses) %in% names(curAnalysis)))
        stop("setCurExperAnalyses: invalid 'analyses' argument")
    for (exper in names(analyses))
        {
        analysis = analyses[exper]
        if (!any(analysis == names(all.phenoData[[exper]])))
            stop("setCurExperAnalyses: no such analysis '", analysis, "' in experiment ", exper)
        curAnalysis[exper] = analysis
        assignGlobal("curAnalysis", curAnalysis)
        }
    }

#######################################################################################
# Get short names for experiments/analyses/subanalyses.
#
# Arguments:
#   exper: vector of experiment names.
#   analysis: vector (same size) of analysis names.  If NULL, use current analysis.
#   subanalysis: vector (same size) of subanalysis names.
#
# Returns: vector of short names for specified experiments/analyses/subanalyses.
# The vector's element names are of the form exper.analysis.sub.
#######################################################################################
getExperAnalysisSub_shortName = function(exper, analysis=NULL, subanalysis)
    {
    if (is.null(analysis))
        analysis = curAnalysis[exper]
    if (length(exper) != length(analysis))
        stop("getExperAnalysisSub_shortName: Expected length of exper and analysis to be identical")
    if (length(exper) != length(subanalysis))
        stop("getExperAnalysisSub_shortName: Expected length of exper and subanalysis to be identical")
    V = c()
    for (i in 1:length(exper))
        V = c(V, all.phenoData[[exper[i]]][[analysis[i]]][[subanalysis[i]]]$shortEAS)
    names(V) = paste(exper, analysis, subanalysis, sep=".")
    return(V)
    }

#######################################################################################
# Get short trait names for experiments/analyses/subanalyses/traits.
#
# Arguments:
#   exper: vector of experiment names.
#   analysis: vector (same size) of analysis names.  If NULL, use current analysis.
#   subanalysis: vector (same size) of subanalysis names.
#   trait: vector (same size) of trait names.
#
# Returns: vector of short names for specified experiments/analyses/subanalyses/traits,
# formed by appending the subanalysis short trait prefix string to the trait name.
# The vector's element names are of the form subanalysis.trait.  Note that the resulting
# names are unique among all subanalyses and traits of that experiment and analysis, but
# not necessarily unique among experiments/analyses.
#######################################################################################
getTrait_shortName = function(exper, analysis=NULL, subanalysis, trait)
    {
    if (is.null(analysis))
        analysis = curAnalysis[exper]
    if (length(exper) != length(analysis))
        stop("getSubanalysisTrait_shortName: Expected length of exper and analysis to be identical")
    if (length(exper) != length(subanalysis))
        stop("getSubanalysisTrait_shortName: Expected length of exper and subanalysis to be identical")
    if (length(exper) != length(trait))
        stop("getSubanalysisTrait_shortName: Expected length of exper and trait to be identical")
    V = c()
    for (i in 1:length(exper))
        V = c(V,
            paste(all.phenoData[[exper[i]]][[analysis[i]]][[subanalysis[i]]]$shortTraitPfx, trait[i], sep=""))
    theNames = paste(subanalysis, trait, sep=".")
    if (!any(duplicated(theNames)))
        names(V) = theNames
    return(V)
    }

#######################################################################################
# Read trait information files for the specified phenotype experiments, subanalyses,
# and traits of the CURRENT ANALYSES into a data frame.
#
# Arguments:
#   subanalysisInExper: see comments at top of file.
#   traits: a vector of trait names (any experiment and subanalysis) for which trait
#       information is desired, or NULL to return data for ALL traits.
#
# Returns: data frame with columns "exper", "analysis", "subanalysis", "trait", "units",
# and "desc".
#######################################################################################
getExperimentTraits = function(subanalysisInExper=NULL, traits=NULL)
    {
    dfExper = getExperimentInfo(names(subanalysisInExper)) # names(NULL) is NULL
    experiments = rownames(dfExper)
    dfTraits = NULL
    for (exper in experiments)
        {
        analysis = curAnalysis[exper]
        analysis.subanalyses = names(all.phenoData[[exper]][[analysis]])
        if (!is.null(subanalysisInExper[[exper]])) # NULL[["x"]] is NULL
            analysis.subanalyses = intersect(analysis.subanalyses, subanalysisInExper[[exper]])
        for (subanalysis in analysis.subanalyses)
            {
            traitsFile = all.phenoData[[exper]][[analysis]][[subanalysis]]$traits
            df = read.table(traitsFile, sep="\t", header=TRUE, row.names=1, stringsAsFactors=FALSE)
            if (!is.null(traits))
                df = df[rownames(df) %in% traits,]
            if (nrow(df) == 0)
                next
            df$exper = exper
            df$analysis = analysis
            df$subanalysis = subanalysis
            df$trait = rownames(df)
            rownames(df) = NULL
            dfTraits = rbind(dfTraits, df[, c("exper", "analysis", "subanalysis", "trait", "units", "desc")])
            }
        }
    # Check to see if any requested traits were not found.
    if (!is.null(traits))
        {
        badTraits = !traits %in% dfTraits$trait
        if (any(badTraits))
            stop("getExperimentTraits: unknown trait name(s): ",
                paste(traits[badTraits], collapse=", "))
        }
    return(dfTraits)
    }

#######################################################################################
# Read trait group files containing group formulas for the specified phenotype
# experiments, subanalyses, and groups of the CURRENT ANALYSES into a data frame.
#
# Arguments:
#   subanalysisInExper: see comments at top of file.
#   groups: a vector of group names (any experiment or subanalysis) for which group
#       formulas are desired, or NULL to return data for ALL groups.
#
# Returns: data frame with columns "group", "exper", "analysis", "subanalysis",
# and "formula".
#######################################################################################
getExperimentGroups = function(subanalysisInExper=NULL, groups=NULL)
    {
    dfExper = getExperimentInfo(names(subanalysisInExper)) # names(NULL) is NULL
    experiments = rownames(dfExper)
    dfGroups = NULL
    for (exper in experiments)
        {
        analysis = curAnalysis[exper]
        analysis.subanalyses = names(all.phenoData[[exper]][[analysis]])
        if (!is.null(subanalysisInExper[[exper]])) # NULL[["x"]] is NULL
            analysis.subanalyses = intersect(analysis.subanalyses, subanalysisInExper[[exper]])
        for (subanalysis in analysis.subanalyses)
            {
            traitGroupsFile = all.phenoData[[exper]][[analysis]][[subanalysis]]$groups
            df = read.table(traitGroupsFile, sep="\t", header=TRUE, stringsAsFactors=FALSE)
            if (!is.null(groups))
                df = df[df$group %in% groups,]
            if (nrow(df) > 0)
                {
                df = data.frame(group=df$group, exper=exper, analysis=analysis, subanalysis=subanalysis,
                    formula=df$formula, row.names=NULL, stringsAsFactors=FALSE)
                dfGroups = rbind(dfGroups, df)
                }
            }
        }
    # Check to see if any requested groups were not found.
    if (!is.null(groups))
        {
        badGroups = !groups %in% dfGroups$group
        if (any(badGroups))
            stop("getExperimentGroups: unknown group name(s): ",
                paste(groups[badGroups], collapse=", "))
        }
    return(dfGroups)
    }

#######################################################################################
# Read the specified traits data files for the specified experiments, subanalyses, and
# traits of the CURRENT ANALYSES into a data frame.
#
# Arguments:
#   subanalysisInExper: see comments at top of file.
#   traits: a vector of trait names (any experiment and subanalysis) for which data is
#       desired, or NULL to return data for all traits.
#   dataKeywords: a vector of one or more data keywords, indicating which data is to
#       be retrieved from the specified experiments, subanalyses, and traits.  See
#       comments at the top of this file describing the data keywords and their data.
#   gtsAsRowNames: if TRUE, the row names of the returned data frame are equal to the
#       genotype column.  If there are any duplicates in the genotype column, an error
#       is given.
#
# Returns: data frame with columns "exper", "analysis", "subanalysis", "trait",
#   "genotype", and column names corresponding to the "dataKeywords" argument.
#   Those column names will not be equal to the "dataKeywords" argument values;
#   each "dataKeyword" value can expand into one or more than one columns of data;
#   see comments in section at the top of this file describing the available
#   phenotype data.
#######################################################################################
getExperimentTraitILdataValues = function(subanalysisInExper=NULL, traits=NULL,
    dataKeywords=data.ILkeywords, gtsAsRowNames=FALSE)
    {
    # Check "dataKeywords" argument.
    badVals = !dataKeywords %in% data.ILkeywords
    if (is.null(dataKeywords) || any(badVals))
        stop("getExperimentTraitILdataValues: unknown 'dataKeywords' argument element(s): ",
            paste(dataKeywords[badVals], collapse=", "))

    # Get analysis type and traits in each experiment.  Side-effect is to check
    # "subanalysisInExper" and "traits" arguments.
    dfTraits = getExperimentTraits(subanalysisInExper, traits)
    experiments = unique(dfTraits$exper)

    # Loop for each experiment name in 'experiments'.
    dfTraitData = NULL
    all.colnames = c("exper", "analysis", "subanalysis", "trait", "genotype")
    for (exper in experiments)
        {
        analysis = curAnalysis[exper]
        analysis.subanalyses = names(all.phenoData[[exper]][[analysis]])
        if (!is.null(subanalysisInExper[[exper]])) # NULL[["x"]] is NULL
            analysis.subanalyses = intersect(analysis.subanalyses, subanalysisInExper[[exper]])

        # Loop for each subanalysis in "exper".
        for (subanalysis in analysis.subanalyses)
            {
            # Get requested traits for this experiment and subanalysis.
            requestedTraits = dfTraits$trait[dfTraits$exper == exper & dfTraits$subanalysis == subanalysis]

            # If there are no traits requested for this experiment and subanalysis, loop next.
            if (length(requestedTraits) == 0)
                next

            # For each requested data keyword, read data files and append data to dfTraitData.
            for (keyword in dataKeywords)
                {
                dataFile = all.phenoData[[exper]][[analysis]][[subanalysis]][[keyword]]
                # A NULL value for the filename means that that experiment does not contain
                # data for that keyword.  Skip it.
                if (is.null(dataFile))
                    next
                # Read the file of trait IL data into data frame df.
                df = read.table(dataFile, sep="\t", header=TRUE, row.names=NULL, stringsAsFactors=FALSE)

                # Several files do not have "trait" and "genotype" columns, but instead, the column
                # names are traits and the row names are genotypes.  Collapse these data frames
                # so there is a "trait" column, a "genotype" column, and a column that is named the
                # same as the data keyword name.  These files can be identified because the first
                # column name is "X" or "row.names", because we should have read the file with row.names=1.
                if (colnames(df)[1] %in% c("X", "row.names"))
                    {
                    df.traits = colnames(df)[-1]
                    df.genotypes = df[,1]
                    dfNew = NULL
                    for (t in df.traits)
                        dfNew = rbind(dfNew, data.frame(trait=t, genotype=df.genotypes,
                            value=df[,t], row.names=NULL, stringsAsFactors=FALSE))
                    colnames(dfNew)[3] = keyword
                    df = dfNew
                    }

                # Select only the rows for the requested traits.
                df = df[df$trait %in% requestedTraits,]
                missingTraits = !requestedTraits %in% unique(df$trait)
                if (any(missingTraits)) 
                    stop("getExperimentTraitILdataValues: trait(s) not found for experiment ", exper,
                        " keyword ", keyword, ": ", paste(requestedTraits[missingTraits], collapse=", "))

                # Add "exper", "analysis", and "subanalysis" as columns in df.
                df$exper = exper
                df$analysis = analysis
                df$subanalysis = subanalysis

                # Each data frame to be added to dfTraitData has columns "exper", "analysis",
                # "subanalysis", "trait", and "genotype", plus one or more other columns.
                # We want to merge based on the "exper", "subanalysis", "trait", and "genotype"
                # columns.  I thought the merge function would work, but it does not, so we
                # have to do this manually.
                if (is.null(dfTraitData))
                    {
                    dfTraitData = df
                    rownames(dfTraitData) = paste(dfTraitData$exper, dfTraitData$subanalysis, dfTraitData$trait, dfTraitData$genotype, sep="_")
                    }
                else
                    {
                    # dfTraitData = merge(dfTraitData, df, all=TRUE)
                    # dfTraitData = merge(dfTraitData, df, by=c("subanalysis", "trait", "genotype"), all=TRUE)

                    # First, add to dfTraitData any new columns present in df.
                    newCols.1 = setdiff(colnames(df), colnames(dfTraitData))
                    dfTraitData[, newCols.1] = NA

                    # Next, determine the column names of the data.  It is not necessarily simply "keyword".
                    dataCols = setdiff(colnames(df), c("exper", "analysis", "subanalysis", "trait", "genotype"))

                    # Next, for those rows that are common between dfTraitData and df,
                    # simply copy the new data from df to dfTraitData.  Common rows are
                    # identified by having the same value for exper, subanalysis, trait,
                    # and genotype.
                    rownames(df) = paste(df$exper, df$subanalysis, df$trait, df$genotype, sep="_")
                    common.L = rownames(df) %in% rownames(dfTraitData)
                    common.names = rownames(df)[common.L]
                    dfTraitData[common.names, dataCols] = df[common.names, dataCols]

                    # Next, add to df any new columns present in dfTraitData.
                    newCols.2 = setdiff(colnames(dfTraitData), colnames(df))
                    df[, newCols.2] = NA

                    # Finally, for those rows that are NOT common, row-bind them to
                    # dfTraitData.
                    dfTraitData = rbind(dfTraitData, df[!common.L, colnames(dfTraitData)])
                    }

                # Accumulate column names that were added.
                all.colnames = c(all.colnames, colnames(df))
                }
            }
        }
    # Try to get columns in order we want, which is the same order that we encountered them.
    all.colnames = unique(all.colnames)
    dfTraitData = dfTraitData[, all.colnames]
    rownames(dfTraitData) = NULL
    # Sort by experiment then subanalysis then trait then genotype.
    ord = order.special(exper=dfTraitData$exper, subanalysis=dfTraitData$subanalysis, trait=dfTraitData$trait,
        genotype=dfTraitData$genotype, lead0="genotype", first=list(genotype=parents))
    dfTraitData = dfTraitData[ord,]
    rownames(dfTraitData) = NULL

    # If requested, make row names be genotype names.
    if (gtsAsRowNames)
        {
        if (any(duplicated(dfTraitData$genotype)))
            stop("getExperimentTraitILdataValues: duplicate genotypes")
        rownames(dfTraitData) = dfTraitData$genotype
        }

    return(dfTraitData)
    }

#######################################################################################
# Read the biorep means data files for the specified experiments and traits of the
# CURRENT ANALYSES into a data frame.
#
# Arguments:
#   subanalysisInExper: see comments at top of file.
#   traits: a vector of trait names (any experiment and subanalysis) for which data is
#       desired, or NULL to retrieve data for all traits.
#
# Returns: data frame with columns "exper", "analysis", "subanalysis", "genotype",
# "biorep", "trait", and "mean".
#
# Note: Currently the only experiment with biorep means is RNAseqRootJun2013.  If any
# other experiment is requested, the returned data frame will simply have no data for
# that experiment.
#######################################################################################
getExperimentTraitILbiorepMeans = function(subanalysisInExper=NULL, traits=NULL)
    {
    # Get traits in each experiment.  Side-effect is to check "subanalysisInExper" and
    # "traits" arguments.
    dfTraits = getExperimentTraits(subanalysisInExper, traits)
    experiments = unique(dfTraits$exper)

    # Loop for each experiment name in 'experiments'.
    dfBiorepMeans = NULL
    for (exper in experiments)
        {
        analysis = curAnalysis[exper]
        analysis.subanalyses = names(all.phenoData[[exper]][[analysis]])
        if (!is.null(subanalysisInExper[[exper]])) # NULL[["x"]] is NULL
            analysis.subanalyses = intersect(analysis.subanalyses, subanalysisInExper[[exper]])

        # Loop for each subanalysis in "exper".
        for (subanalysis in analysis.subanalyses)
            {
            # Get requested traits for this experiment.
            requestedTraits = dfTraits$trait[dfTraits$exper == exper & dfTraits$subanalysis == subanalysis]

            # If there are no traits requested for this experiment, loop next.
            if (length(requestedTraits) == 0)
                next

            # Read biorep means data file and append data to dfBiorepMeans.
            dataFile = all.phenoData[[exper]][[analysis]][[subanalysis]][["biorepMeans"]]
            # A NULL value for the filename means that that experiment does not contain
            # data for that keyword.  Skip it.
            if (is.null(dataFile))
                next
            # Read the file.
            df = read.table(dataFile, sep="\t", header=TRUE, row.names=NULL, stringsAsFactors=FALSE)

            # Check for missing traits.
            missingTraits = !requestedTraits %in% colnames(df)
            if (any(missingTraits)) 
                stop("getExperimentTraitILbiorepMeans: trait(s) not found for experiment ", exper,
                    " : ", paste(requestedTraits[missingTraits], collapse=", "))

            # For each requested trait, get the trait column along with genotype and biorep,
            # and add the data to dfBiorepMeans.
            for (trait in requestedTraits)
                {
                df2 = df[, c("genotype", "biorep", trait)]
                colnames(df2)[3] = "mean"

                # Add "exper", "analysis", "subanalysis", and "trait" as columns in df2.
                df2$exper = exper
                df2$analysis = analysis
                df2$subanalysis = subanalysis
                df2$trait = trait

                # Put columns in desired order and append to dfBiorepMeans.
                dfBiorepMeans = rbind(dfBiorepMeans,
                    df2[, c("exper", "analysis", "subanalysis", "genotype", "biorep", "trait", "mean")])
                }
            }
        }

    # Sort by experiment then subanalysis then trait then genotype then biorep.
    ord = order.special(exper=dfBiorepMeans$exper, subanalysis=dfBiorepMeans$subanalysis,
        trait=dfBiorepMeans$trait, genotype=dfBiorepMeans$genotype, biorep=dfBiorepMeans$biorep,
        lead0="genotype", first=list(genotype=parents))
    dfBiorepMeans = dfBiorepMeans[ord,]
    rownames(dfBiorepMeans) = NULL
    return(dfBiorepMeans)
    }

#######################################################################################
# Read trait groups files for the specified phenotype experiments and groups of the
# CURRENT ANALYSES and then determine which traits participate in the trait formulas
# for each group.
#
# Arguments:
#   subanalysisInExper: see comments at top of file.
#   groups: a vector of group names (any experiment and subanalysis) for which traits
#       are desired, or NULL to return data for ALL groups.
#
# Returns: data frame with columns "exper", "analysis", "subanalysis", "group", and
# "trait".
#######################################################################################
getTraitsInGroups = function(subanalysisInExper=NULL, groups=NULL)
    {
    # Get groups for the experiments.
    dfGroups = getExperimentGroups(subanalysisInExper, groups)

    # Loop for each group in dfGroups, parse its group formula, and accumulate
    # trait names in data frame dfGroupTraits.
    dfGroupTraits = NULL
    if (nrow(dfGroups) > 0)
        {
        for (i in 1:nrow(dfGroups))
            {
            group = dfGroups$group[i]

            # Get the formula for this group and parse it to create a postfix vector.
            postfix = parseFormula(dfGroups[i, "formula"])

            # Extract trait names from postfix vector.
            traits = postfix[!postfix %in% c("!", "&", "|", "(", ")")]

            # Add the data to dfGroupTraits.
            dfGroupTraits = rbind(dfGroupTraits,
                data.frame(dfGroups[i, c("exper", "analysis", "subanalysis", "group")],
                    trait=traits, row.names=NULL, stringsAsFactors=FALSE))
            }
        }
    return(dfGroupTraits)
    }

#######################################################################################
# Read trait groups files for the specified experiments, subanalyses, and groups, and
# read the trait data values for all traits of the specified experiments/subanalyses
# and of the CURRENT ANALYSES.
# Then, for the specified groups, parse their group formulas and examine the requested
# trait data value given by traitDataCol (e.g. mean, q-value, etc.), and based on the
# formula, gather a list of ILs satisfying the trait value requirements as specified
# in the arguments traitDataCol, thresholdOp, and thresholdVal.  Those ILs are
# returned.  See examples below.
#
# Group formulas take a simple form, but provide great flexibility in creating
# sets of ILs with significant traits.  A formula applies to the traits in a
# single experiment/analysis/subanalysis. The group formula language is as
# follows:
#
#   GROUP FORMULA :=    TERM
#
#   TERM :=             TRAIT               or
#                       TRAIT | TERM        or
#                       TRAIT & TERM        or
#                       ! TERM              or
#                       ( TERM )
#
#   TRAIT :=            name of any trait in the experiment
# 
# Here are some example group formulas:
#                       gravity_d5
#                       gravity_d5 | gravity_d7
#                       avgAbsAng & tipAng
#                       avgAbsAng & !tipAng
#                       !area_root
#                       avgAbsAng & !(tipAng | wholeAng)
#
# The operators have the following meanings:
#       |       "or" of trait ILs; union of ILs specified by the terms on each side of the |.
#       &       "and" of trait ILs; intersection of ILs specified by the terms on each side of the &.
#       !       "not" of trait ILs; set difference of all trait ILs minus ILs specified by term after !.
# The & and | operators have equal precedence, so if you want "a | (b & c)" you
# must use the parentheses and cannot write it as "a | b & c".
#
# Where do the ILs come from, if only trait names are specified?  They are determined
# using the trait name and the arguments traitDataCol, thresholdOp, and thresholdVal.
# These arguments tell which trait value to look at (e.g. mean or q-value) and what
# threshold qualifies an IL to be included as "significant".  All ILs are examined
# for that data value of that trait, and if it meets the threshold, the IL is deemed
# "significant".  See examples below.
#
# Arguments:
#   subanalysisInExper: see comments at top of file.
#   groups: a vector of group names (any experiment and subanalysis) for which
#       group formulas are to be processed, NULL for all groups.
#   traitDataCol: name of a data column in the data frame returned by
#       getExperimentTraitILdataValues(), whose value in ILs is compared to
#       thresholdVal to decide which ILs are significant.
#   thresholdOp: a comparison operator, one of ("=", "!=", "<", "<=", ">", ">="),
#       defining the type of comparison to make between:
#       1) (left of thresholdOp): a value named by traitDataCol which is derived
#           from an IL's trait comparison to M82 (e.g. traitDataCol might be "mean"
#           or "qvalue")
#       2) (right of thresholdOp): the value given by the thresholdVal argument.
#   thresholdVal: the value to compare to the IL's trait results traitDataCol value.
#       It must be the same R class as the traitDataCol column.  thresholdVal may be
#       NA, in which case the comparison is to the value of the M82 genotype in
#       the same experiment/trait.
#   FURTHER INFO ON traitDataCol, thresholdOp, and thresholdVal: these may actually
#       be vectors larger than 1 (all the same size), or in the case of thresholdVal,
#       a list (because different thresholdVals may be different R classes so a
#       vector can't hold them), in which case the conditions specified are applied
#       sequentially and only if all succeed is the IL deemed "significant".
#       See examples below.
#
# Returns: data frame with columns "exper", "analysis", "subanalysis", "group",
# "formula", "condition", and "IL", giving all the ILs satisfying the formulas
# for the specified experiments, subanalyses, and groups of the CURRENT analyses.
#
# Examples:
#
#   1.  If traitDataCol="qvalue", thresholdOp="<=", and thresholdVal=0.01,
#       and if one of the specified experiments/groups had the formula:
#                   "avgAbsAng | tipAng"
#       then the ILs with qvalue <= 0.01 for either the avgAbsAng OR tipAng traits
#       would be included among the returned data frame rows.
#
#   2.  If traitDataCol=c("qvalue", "mean"), thresholdOp=c("<=", ">"), and
#       thresholdVal=c(0.01, NA), and if experiments were NULL and groups were
#       "cells_xylem", and if group "cells_xylem" had the formula:
#                   "cells_xylem" (i.e. group is named the same as its sole trait)
#       then the ILs with qvalue <= 0.01 and mean > M82_mean for the cells_xylem
#       trait would be included among the returned data frame rows.
#######################################################################################
getILsInGroups = function(subanalysisInExper=NULL, groups=NULL,
    traitDataCol="qvalue", thresholdOp="<=", thresholdVal=0.01)
    {
    # Check that traitDataCol, thresholdOp, and thresholdVal are the same length.
    if (length(traitDataCol) != length(thresholdOp))
        stop("getILsInGroups: traitDataCol and thresholdOp lengths differ")
    if (length(traitDataCol) != length(thresholdVal))
        stop("getILsInGroups: traitDataCol and thresholdVal lengths differ")

    # Check thresholdOp argument.
    thresholdOps = c("=", "!=", "<", "<=", ">", ">=")
    badThresholdOps = !thresholdOp %in% thresholdOps
    if (any(badThresholdOps))
        stop("getILsInGroups: invalid thresholdOp argument(s): ",
            paste(thresholdOp[badThresholdOps], collapse=","))

    # Get groups for the experiments.
    dfGroups = getExperimentGroups(subanalysisInExper, groups)

    # Get all phenotype data for the experiments.
    dfTraitData = getExperimentTraitILdataValues(subanalysisInExper)

    # Make sure each traitDataCol is among the column names.
    badTraitDataCols = !traitDataCol %in% colnames(dfTraitData)
    if (any(badTraitDataCols))
        stop("getILsInGroups: invalid traitDataCol value(s): ",
            paste(traitDataCol[badTraitDataCols], collapse=","))

    # Make sure the class of each thresholdVal matches that of the corresponding
    # traitDataCol column.
    for (i in 1:length(thresholdVal))
        {
        tv = thresholdVal[[i]] # Double bracket in case thresholdVal is a list.
        if (!is.na(tv))
            {
            col = traitDataCol[i]
            if (is.numeric(dfTraitData[, col]) && !is.numeric(tv))
                stop("getILsInGroups: expected numeric thresholdVal argument for: ", tv)
            else if (is.character(dfTraitData[, col]) && !is.character(tv))
                stop("getILsInGroups: expected character thresholdVal argument for: ", tv)
            else if (is.logical(dfTraitData[, col]) && !is.logical(tv))
                stop("getILsInGroups: expected logical thresholdVal argument for: ", tv)
            }
        }

    # Create a string giving the complete set of conditions defined by
    # traitDataCol, thresholdOp, and thresholdVal, for inclusion in output
    # data frame column "condition".
    condition = paste(traitDataCol, thresholdOp, thresholdVal, sep=" ", collapse=" && ")

    # Loop for each row of dfGroups, and process its group formula.  Accumulate
    # significant ILs in data frame dfILs.
    dfILs = NULL
    if (nrow(dfGroups) > 0)
        {
        for (i in 1:nrow(dfGroups))
            {
            exper = dfGroups$exper[i]
            analysis = dfGroups$analysis[i]
            subanalysis = dfGroups$subanalysis[i]
            group = dfGroups$group[i]
            form = dfGroups$formula[i]

            # Parse the group formula to create a postfix vector.
            postfix = parseFormula(form)

            # Get the subset of trait data for the experiment/subanalysis that this group is a member of.
            dfTraitData.thisGroup = dfTraitData[dfTraitData$exper == exper & dfTraitData$subanalysis == subanalysis,]

            # Get the set of all traits for the experiment/subanalysis that this group is a member of.
            traits = unique(dfTraitData.thisGroup$trait)

            # Get the set of all ILs (plus parents) for the experiment/subanalysis
            # that this group is a member of.
            AllILs = unique(dfTraitData.thisGroup$genotype)

            # Evaluate the formula to get the list of ILs.  The skeleton of this is
            # taken from testEvaluateFormula(), with comments removed.  Note that
            # stk[] must be a list rather than vector, because its elements are
            # vectors (of ILs).
            stk = list()
            while (length(postfix) > 0)
                {
                token = postfix[1]
                postfix = postfix[-1]
                if (grepl("^[A-Za-z0-9_\\.]+$", token))
                    {
                    # TRAIT token.  Here is where we look up the ILs significant for the
                    # specified value in the trait given by the trait token.
                    trait = token
                    if (!trait %in% traits) stop("getILsInGroups: invalid trait name: ", trait, " in formula: ", form)
                    df = dfTraitData.thisGroup[dfTraitData.thisGroup$trait == trait,]

                    # Loop for each thresholdVal and test it.  Start with all ILs, and
                    # at each test, intersect the ILs passing the test with the result
                    # ILs.
                    resultILs = AllILs
                    for (i in 1:length(thresholdVal))
                        {
                        col = traitDataCol[i]
                        op = thresholdOp[i]
                        tv = thresholdVal[[i]] # Double bracket in case thresholdVal is a list.
                        if (is.na(tv))
                            {
                            tv = df[df$genotype == "M82", col]
                            if (length(tv) == 0 || is.na(tv))
                                stop("getILsInGroups: missing M82 trait value: ", col, " in trait: ", trait)
                            }
                        vals = df[, col] # Some vals may be NA
                        if (op == "=")
                            isSig = (vals == tv)
                        else if (op == "!=")
                            isSig = (vals != tv)
                        else if (op == "<")
                            isSig = (vals < tv)
                        else if (op == "<=")
                            isSig = (vals <= tv)
                        else if (op == ">")
                            isSig = (vals > tv)
                        else # (op == ">=")
                            isSig = (vals >= tv)
                        ILs = unique(df$genotype[isSig])
                        # If any vals were NA, we will have NA in the ILs list.  Remove it.
                        # Also remove the parents, which can also be present.
                        ILs = setdiff(ILs, c(parents, NA))
                        # Intersect the ILs with the current result ILs.
                        resultILs = intersect(resultILs, ILs)
                        # If any vals were NA, those were ILs that were not ascertained for the trait.
                        # They should be considered neither significant nor non-significant.  To
                        # accomplish this, remove all such ILs from both resultILs and AllILs.  We
                        # don't want the "!" operator below to reintroduce any of those ILs.
                        NA_ILs = unique(df$genotype[is.na(vals)])
                        resultILs = setdiff(resultILs, NA_ILs)
                        AllILs = setdiff(AllILs, NA_ILs)
                        }
                    #cat("ILs for trait ", trait, " are ", paste(resultILs, collapse=","), "\n\n")
                    stk = c(stk, list(resultILs))
                    }
                else if (token == "|")
                    {
                    # | is union of two sets of ILs
                    N = length(stk)
                    ILs1 = stk[[N-1]]
                    ILs2 = stk[[N]]
                    stk = stk[-N]
                    ILs = union(ILs1, ILs2)
                    stk[[N-1]] = ILs
                    }
                else if (token == "&")
                    {
                    # & is intersection of two sets of ILs
                    N = length(stk)
                    ILs1 = stk[[N-1]]
                    ILs2 = stk[[N]]
                    stk = stk[-N]
                    ILs = intersect(ILs1, ILs2)
                    stk[[N-1]] = ILs
                    }
                else # (token == "!")
                    {
                    # ! is difference of ALL ILs and the ILs on the stack
                    N = length(stk)
                    ILs = stk[[N]]
                    ILs = setdiff(AllILs, ILs)
                    stk[[N]] = ILs
                    }
                }
            #   Else that value is the result of the calculation.
            ILs = stk[[1]]
            if (length(ILs) > 0)
                dfILs = rbind(dfILs, data.frame(exper=exper, analysis=analysis,
                    subanalysis=subanalysis, group=group, formula=form,
                    condition=condition, IL=ILs,
                    row.names=NULL, stringsAsFactors=FALSE))
            }
        }
    if (is.null(dfILs))
        dfILs = data.frame(exper=c(), analysis=c(), subanalysis=c(), group=c(),
            formula=c(), condition=c(), IL=c(), row.names=NULL, stringsAsFactors=FALSE)
    return(dfILs)
    }

#######################################################################################
# Parse a group formula into postfix notation.
#
# Arguments:
#   form: a group formula character string (see getILsInGroups() comments).
#
# Returns: a vector containing a postfix representation of the formula that can be
#   evaluated easily using a stack.  Each vector element is either a trait name or
#   an operator ("&", "|", "!").  See function testEvaluateFormula() for code that
#   does a test evaluation of the postfix vector.
#
# Example: given the formula "a & (b | !c) | !(!d & e)", the postfix representation is:
#       a, b, c, !, |, &, d, !, e, &, !, |
# Starting from the first token (element in returned vector), the expression is
# evaluated using a stack, each element of which can hold a vector of ILs.
# Each token is processed as follows:
#       Token Type      Action
#       ----------      ------------------------------------------------------------------------------
#       trait name      evaluate sig ILs of the trait and push a vector of the ILs onto the stack.
#           !           x <- pop_stack;  setdiff(allILs, x) -> push_stack
#           &           x2 <- pop_stack;  x1 <- pop_stack;  intersect(x1, x2) -> push_stack
#           |           x2 <- pop_stack;  x1 <- pop_stack;  union(x1, x2) -> push_stack
#######################################################################################
parseFormula = function(form)
    {
    # Let's make sure all operators and parentheses have spaces around them, then
    # split the formula on space.
    f = form
    f = gsub("&", " & ", f, fixed=TRUE)
    f = gsub("|", " | ", f, fixed=TRUE)
    f = gsub("!", " ! ", f, fixed=TRUE)
    f = gsub("(", " ( ", f, fixed=TRUE)
    f = gsub(")", " ) ", f, fixed=TRUE)
    f = unlist(strsplit(f, " +"))
    f = f[f != ""]
    if (length(f) == 0) stop("parseFormula: empty formula: ", form)

    # We will use two vectors, postfix and stk, to parse the expression.  We will put
    # parsed tokens awaiting an additional token (such as "a |" or "!") onto stk.
    # When we have the additional token, remove the other(s) from the stack and
    # put them into vector postfix in the correct order for evaluation.
    postfix = c()
    stk = c()

    # Define what are operators, what is their precedence, and are they left- or right-associative.
    ops = c("|", "&", "!")
    opPrecedence = c(`|` = 1, `&` = 1, `!` = 2)
    opLeftAssociative = c(`|` = TRUE, `&` = TRUE, `!` = FALSE)

    # Use a simplified version of the Shunting-yard algorithm to convert the formula,
    # which is in infix notation, into postfix notation.  The algorithm is as follows,
    # paraphrased from http://en.wikipedia.org/wiki/Shunting-yard_algorithm
    #   While there are input tokens in f[] to be processed:
    #       Get a token from f[].
    #       If the token is a TRAIT name, append it to postfix[].
    #       If the token is an operator, o1, then:
    #           while there is an operator token, o2, at the top of stk[] and either:
    #               o1 precedence is less than that of o2, or
    #               o1 is left associative and has precedence equal to that of o2,
    #                   pop o2 off stk[] and append it to postfix[].
    #           push o1 onto stk[].
    #       If the token is a left parenthesis, then push it onto stk[].
    #       If the token is a right parenthesis, then:
    #           Until the token at the top of stk[] is a left parenthesis
    #               pop operators off stk[] and append to postfix[].
    #           If stk[] runs out without finding a left parenthesis, there are mismatched parentheses.
    #           Pop the left parenthesis off stk[] and discard it.
    #       If the token is some other unrecognized value, there is a bad token.
    #   When there are no more tokens to read and there are still operator tokens on stk[]
    #       If token on the top of stk[] is a left parenthesis, there are mismatched parentheses.
    #       Pop the operator off stk[] and append to postfix[].
    #   Exit.

    # Now execute the above algorithm.
    #   While there are input tokens in f[] to be processed:
    i = 1
    while (i <= length(f))
        {
        # Get a token from f[].
        token = f[i]
        i = i+1
        #  If the token is a TRAIT name, append it to postfix[].
        if (grepl("^[A-Za-z0-9_\\.]+$", token))
             postfix = c(postfix, token)
        # If the token is an operator, o1, then:
        else if (token %in% ops)
            {
            # while there is an operator token, o2, at the top of stk[] and either:
            #   o1 precedence is less than that of o2, or
            #   o1 is left associative and has precedence equal to that of o2,
            N = length(stk)
            while (N > 0 && stk[N] %in% ops && (
                (opPrecedence[token] < opPrecedence[stk[N]]) ||
                ((opPrecedence[token] == opPrecedence[stk[N]]) && opLeftAssociative[token])))
                {
                # pop o2 off stk[] and append it to postfix[].
                postfix = c(postfix, stk[N])
                stk = stk[-N]
                N = N-1
                }
            # push o1 onto stk[].
            stk = c(stk, token)
            }
        # If the token is a left parenthesis, then push it onto stk[].
        else if (token == "(")
            stk = c(stk, token)
        # If the token is a right parenthesis, then:
        else if (token == ")")
            {
            # Until the token at the top of stk[] is a left parenthesis
            N = length(stk)
            while (N > 0 && stk[N] != "(")
                {
                # pop operators off stk[] and append to postfix[].
                postfix = c(postfix, stk[N])
                stk = stk[-N]
                N = N-1
                }
            # If stk[] runs out without finding a left parenthesis, there are mismatched parentheses.
            if (length(stk) == 0) stop("parseFormula: missing left parenthesis: ", form)
            # Pop the left parenthesis off stk[] and discard it.
            stk = stk[-N]
            }
        # If the token is some other unrecognized value, there is a bad token.
        else stop("parseFormula: unknown value: ", token, " in formula: ", form)
        }
    # When there are no more tokens to read and there are still operator tokens on stk[]
    N = length(stk)
    while (N > 0)
        {
        # If token on the top of stk[] is a left parenthesis, there are mismatched parentheses.
        if (stk[N] == "(") stop("parseFormula: missing right parenthesis: ", form)
        # Pop the operator off stk[] and append to postfix[].
        postfix = c(postfix, stk[N])
        stk = stk[-N]
        N = N-1
        }
    
    #   Exit.

    # Perform a test evaluation to check for errors.
    testEvaluateFormula(postfix, form)

    return(postfix)
    }

#######################################################################################
# Perform a test evaluation of a group formula in postfix notation, as returned by
# parseFormula().
#
# Arguments:
#   postfix: a vector of trait names and operators in postfix format.
#   form: formula that the postfix vector was generated from.
#
# If any errors occur during evaluation of the postfix expression, stop() is called.
#
# To implement a group formula evaluator, copy this code and modify it.
#
# Algorithm:
# From Wikipedia (http://en.wikipedia.org/wiki/Reverse_Polish_notation#Postfix_algorithm):
# The algorithm for evaluating any postfix expression is fairly straightforward:
#   While there are input tokens left in postfix
#       Get the next token
#       If the token is a trait name
#           Push it onto the stack.
#       Otherwise, the token is an operator.
#           It is known a priori that the operator takes n arguments.
#           If there are fewer than n values on the stack
#               (Error) The user has not input sufficient values in the expression.
#           Else Pop the top n values from the stack.
#           Evaluate the operator, with the values as arguments.
#           Push the returned results, if any, back onto the stack.
#   If there is not exactly one value on the stack
#       (Error) The user input has too many values.
#   Else that value is the result of the calculation.
#######################################################################################
testEvaluateFormula = function(postfix, form)
    {
    # We use a stack, which will be vector "stk", for evaluation.
    stk = c()
    # Number of arguments taken by each operator.
    opNumArgs = c(`|` = 2, `&` = 2, `!` = 1)
    # Execute the algorithm.
    #   While there are input tokens left in postfix
    while (length(postfix) > 0)
        {
        #     Get the next token
        token = postfix[1]
        postfix = postfix[-1]
        #     If the token is a trait name
        if (grepl("^[A-Za-z0-9_\\.]+$", token))
            #         Push it onto the stack.
            stk = c(stk, token)
        #     Otherwise, the token is an operator.
        else
            {
            #         It is known a priori that the operator takes n arguments.
            n = opNumArgs[token]
            #         If there are fewer than n values on the stack
            N = length(stk)
            if (N < n)
            #             (Error) The user has not input sufficient values in the expression.
                stop("testEvaluateFormula: operator ", token, " used incorrectly in group formula: ", form)
            #         Else Pop the top n values from the stack.
            stk = stk[-((N-n+1):N)]
            #         Evaluate the operator, with the values as arguments.
            # (test evaluation, we don't really evaluate the operator)
            #         Push the returned results, if any, back onto the stack.
            stk = c(stk, "x") # Fake the result
            }
        }
    #   If there is not exactly one value on the stack
    if (length(stk) != 1)
    #       (Error) The user input has too many values.
        stop("testEvaluateFormula: invalid syntax in group formula: ", form)
    #   Else that value is the result of the calculation.
    }

#######################################################################################
# End of file.
#######################################################################################
Sourced.Include_phenoAnalysis = TRUE
cat("  Include_phenoAnalysis.R included\n")
